from typing import Iterable

from sapiopylib.rest.User import SapioUser
from sapiopylib.rest.pojo.CustomReport import CustomReportCriteria
from sapiopylib.rest.pojo.webhook.WebhookContext import SapioWebhookContext
from sapiopylib.rest.utils.recordmodel.RecordModelWrapper import WrappedType

from sapiopycommons.customreport.custom_report_builder import CustomReportBuilder
from sapiopycommons.datatype.data_fields import ProcessQueueItemFields, SystemFields, ProcessWorkflowTrackingFields
from sapiopycommons.general.aliases import UserIdentifier, AliasUtil, SapioRecord
from sapiopycommons.general.custom_report_util import CustomReportUtil
from sapiopycommons.general.exceptions import SapioException
from sapiopycommons.general.time_util import TimeUtil
from sapiopycommons.recordmodel.record_handler import RecordHandler
from sapiopycommons.webhook.webhook_context import ProcessQueueContext


class QueueItemReportCriteria:
    """
    Queue item report criteria is used to restrict the results of searches for queue item records.
    """
    process_names: list[str] | None
    not_process_names: list[str] | None
    step_names: list[str] | None
    not_step_names: list[str] | None
    data_type_names: list[str] | None
    not_data_type_names: list[str] | None
    data_record_ids: list[int] | None
    not_data_record_ids: list[int] | None
    assigned_to: list[str] | None
    not_assigned_to: list[str] | None
    launched_after: int | None
    launched_before: int | None
    scheduled_after: int | None
    scheduled_before: int | None
    shown_in_queue: bool | None
    has_experiment: bool | None

    def __init__(self, *,
                 process_names: list[str] | None = None,
                 not_process_names: list[str] | None = None,
                 step_names: list[str] | None = None,
                 not_step_names: list[str] | None = None,
                 data_type_names: list[str] | None = None,
                 not_data_type_names: list[str] | None = None,
                 data_record_ids: list[int] | None = None,
                 not_data_record_ids: list[int] | None = None,
                 assigned_to: list[str] | None = None,
                 not_assigned_to: list[str] | None = None,
                 launched_after: int | None = None,
                 launched_before: int | None = None,
                 scheduled_after: int | None = None,
                 scheduled_before: int | None = None,
                 shown_in_queue: bool | None = None,
                 has_experiment: bool | None = None):
        """
        :param process_names: The allowed process name(s).
        :param not_process_names: The disallowed process name(s).
        :param step_names: The allowed step name(s).
        :param not_step_names: The disallowed step name(s).
        :param data_type_names: The allowed data type name(s).
        :param not_data_type_names: The disallowed dta type name(s).
        :param data_record_ids: The allowed record ID(s).
        :param not_data_record_ids: The disallowed record ID(s).
        :param assigned_to: The allowed username(s) of the user(s) that the queue items are assigned to.
        :param not_assigned_to: The disallowed username(s) of the user(s) that the queue items are assigned to.
        :param launched_after: A timestamp after which the queue item was launched.
        :param launched_before: A timestamp before which the queue item was launched.
        :param scheduled_after: A timestamp after which the queue item was scheduled.
        :param scheduled_before: A timestamp before which the queue item was scheduled.
        :param shown_in_queue: Whether the queue item is currently being shown in a queue.
        :param has_experiment: Whether the queue item is linked to an experiment record.
        """
        self.process_names = process_names
        self.not_process_names = not_process_names
        self.step_names = step_names
        self.not_step_names = not_step_names
        self.data_type_names = data_type_names
        self.not_data_type_names = not_data_type_names
        self.data_record_ids = data_record_ids
        self.not_data_record_ids = not_data_record_ids
        self.assigned_to = assigned_to
        self.not_assigned_to = not_assigned_to
        self.launched_after = launched_after
        self.launched_before = launched_before
        self.scheduled_after = scheduled_after
        self.scheduled_before = scheduled_before
        self.shown_in_queue = shown_in_queue
        self.has_experiment = has_experiment


class QueueItemHandler:
    """
    A class used for handling the display of records in custom process queues, which are controlled in the system by
    ProcessQueueItem data types.
    """
    user: SapioUser
    context: ProcessQueueContext | None
    rec_handler: RecordHandler

    def __init__(self, context: UserIdentifier):
        """
        :param context: The current webhook context or a user object to send requests from.
        """
        self.user = AliasUtil.to_sapio_user(context)
        self.rec_handler = RecordHandler(self.user)
        if isinstance(context, SapioWebhookContext):
            self.context = ProcessQueueContext(context)
        else:
            self.context = None

    def get_process_queue_items_from_context(self, wrapper: type[WrappedType],
                                             context: SapioWebhookContext | ProcessQueueContext | None = None) \
            -> list[WrappedType]:
        """
        When you launch records from a custom process queue, the process queue items related to the selected records
        are provided as record IDs to the process queue context. Using these record IDs, query for the queue item
        records and wrap them as record models.

        :param wrapper: The record model wrapper for the process queue items.
        :param context: If this handler was not initialized with a context object, or you wish to retrieve
            data from a different context object than the initializing context, then provide the context to retrieve the
            record IDs from.
        :return: The process queue items corresponding to the record IDs from the context wrapped as record models.
        """
        if context is None and self.context is not None:
            record_ids: list[int] = self.context.process_queue_item_record_ids
        elif context is not None:
            if isinstance(context, SapioWebhookContext):
                record_ids: list[int] = ProcessQueueContext(context).process_queue_item_record_ids
            else:
                record_ids: list[int] = context.process_queue_item_record_ids
        else:
            raise SapioException("A context object must be provided to this function call, as the handler "
                                 "was not initialized with one.")
        return self.rec_handler.query_models_by_id(wrapper, record_ids)

    def map_records_to_queue_items(self, records: Iterable[SapioRecord], queue_items: Iterable[SapioRecord]) \
            -> dict[SapioRecord, list[SapioRecord]]:
        """
        Given a list of records and a list of queue items, create a dictionary mapping the records to the queue items
        that refer to them.

        :param records: The records to map to the queue items.
        :param queue_items: The queue items to map to the records.
        :return: A dictionary of record to the queue items that refer to them. Input queue items that don't refer to
            any provided records will not be in this dictionary.
        """
        ret_val: dict[SapioRecord, list[SapioRecord]] = {}
        id_to_queue_items: dict[int, list[SapioRecord]] = self.rec_handler.map_by_field(queue_items,
                                                                                        ProcessQueueItemFields.DATA_RECORD_ID__FIELD)
        id_to_record: dict[int, SapioRecord] = self.rec_handler.map_by_id(records)
        for record_id, record in id_to_record.items():
            ret_val[record] = id_to_queue_items[record_id]
        return ret_val

    def map_queue_items_to_records(self, queue_items: Iterable[SapioRecord], records: Iterable[SapioRecord]) \
            -> dict[SapioRecord, SapioRecord]:
        """
        Given a list of queue items and a list of records, create a dictionary mapping the queue items to the record
        that they refer to.

        :param queue_items: The queue items to map to the records.
        :param records: The records to map to the queue items.
        :return: A dictionary of queue items to the records that the refer to. Input record that aren't referred to by
            any provided queue items will not be in this dictionary.
        """
        ret_val: dict[SapioRecord, SapioRecord] = {}
        id_to_queue_items: dict[int, list[SapioRecord]] = self.rec_handler.map_by_field(queue_items,
                                                                                        ProcessQueueItemFields.DATA_RECORD_ID__FIELD)
        id_to_record: dict[int, SapioRecord] = self.rec_handler.map_by_id(records)
        for record_id, queue_items in id_to_queue_items.items():
            record: SapioRecord = id_to_record[record_id]
            for queue_item in queue_items:
                ret_val[queue_item] = record
        return ret_val

    def get_queue_items_from_report(self, wrapper: type[WrappedType], criteria: QueueItemReportCriteria) \
            -> list[WrappedType]:
        """
        Run a custom report that retrieves every queue item in the system for the given search criteria.

        :param wrapper: The record model wrapper for the process queue items.
        :param criteria: The search criteria to query for queue items with.
        :return: A list of every queue item in the system that matches the search criteria.
        """
        report = self.build_queue_item_report(criteria)
        return self.rec_handler.query_models_by_report(wrapper, report)

    def get_records_from_item_report(self, wrapper: type[WrappedType],
                                     criteria: QueueItemReportCriteria = QueueItemReportCriteria()) -> list[WrappedType]:
        """
        Run a custom report that retrieves for queue items that match the given search criteria, then query for the
        data records that those queue items refer to.

        :param wrapper: The record model wrapper for the records being queried for.
        :param criteria: Additional search criteria to filter the results. This function forces the data_type_names
            parameter of the criteria to match the data type of the given record model wrapper.
        :return: A list of all records related to the queue items in the system that match the search criteria.
        """
        # Don't try to query for process queue items that don't match the data type of this wrapper.
        criteria.data_type_names = [wrapper.get_wrapper_data_type_name()]
        criteria.not_data_type_names = None
        report = self.build_queue_item_report(criteria)
        record_ids: list[int] = [x[ProcessQueueItemFields.DATA_RECORD_ID__FIELD.field_name]
                                 for x in CustomReportUtil.run_custom_report(self.user, report)]
        return self.rec_handler.query_models_by_id(wrapper, record_ids)

    def get_queue_items_for_records(self, records: Iterable[SapioRecord], wrapper: type[WrappedType],
                                    criteria: QueueItemReportCriteria = QueueItemReportCriteria()) \
            -> dict[SapioRecord, list[WrappedType]]:
        """
        Given a list of records, query the system for every process queue item that refers to those records and matches
        the provided search criteria.

        :param records: The queued records to query for the process queue items of.
        :param wrapper: The record model wrapper for the returned process queue item records.
        :param criteria: Additional search criteria to filter the results. This function forces the data_record_ids and
            data_type_names parameters on the criteria to match the given records.
        :return: A dictionary mapping the input records to a list of the process queue items that refer to them. If a
            record does not have any queue items that refer to it that match the given search criteria, then it will
            map to an empty list.
        """
        # Query for only those process queue items that have a record ID from the provided records.
        criteria.data_record_ids = AliasUtil.to_record_ids(records)
        criteria.not_data_record_ids = None
        criteria.data_type_names = AliasUtil.to_data_type_names(records)
        criteria.not_data_type_names = None
        items: list[WrappedType] = self.get_queue_items_from_report(wrapper, criteria)
        return self.map_records_to_queue_items(records, items)

    def get_records_for_queue_items(self, queue_items: Iterable[SapioRecord], wrapper: type[WrappedType]) \
            -> dict[SapioRecord, WrappedType]:
        """
        Given a list of process queue items, query the system for the records that those queue items refer to.

        :param queue_items: The process queue items to query for the referenced records of.
        :param wrapper: The record model wrapper for the records being queried.
        :return: A dictionary mapping the input process queue items to the record tht they refer to.
        """
        record_ids: set[int] = {x.get_field_value(ProcessQueueItemFields.DATA_RECORD_ID__FIELD) for x in queue_items}
        records: list[WrappedType] = self.rec_handler.query_models_by_id(wrapper, record_ids)
        return self.map_queue_items_to_records(queue_items, records)

    def queue_records_for_process(self, records: Iterable[SapioRecord], process: str, step: str,
                                  wrapper: type[WrappedType]) -> dict[SapioRecord, WrappedType]:
        """
        Given a list of records, create process queue item records for them at the provided process and step names.
        You must store and commit using the record model manager in order for these changes to take effect.

        :param records: The records to create process queue items for.
        :param wrapper: The record model wrapper for the process queue items being created.
        :param process: The name of the process to queue for.
        :param step: The name of the step in the above process to queue for. This is the "Workflow Name" field of the
            Process Workflow record corresponding to the step you want to assign these records to. For steps that
            launch an experiment, this is the name of the template that will be launched. For the other types of custom
            process steps, this is the "Workflow Name" as defined in the process manager config.
        :return: A dictionary mapping each input record to the newly created process queue item for that record.
        """
        ret_val: dict[SapioRecord, WrappedType] = {}
        for record in records:
            item = self.rec_handler.add_model(wrapper)
            item.set_field_values({
                ProcessQueueItemFields.PROCESS_HEADER_NAME__FIELD.field_name: process,
                ProcessQueueItemFields.WORKFLOW_HEADER_NAME__FIELD.field_name: step,
                ProcessQueueItemFields.SHOW_IN_QUEUE__FIELD.field_name: True,
                ProcessQueueItemFields.SCHEDULED_DATE__FIELD.field_name: TimeUtil.now_in_millis(),
                ProcessQueueItemFields.DATA_RECORD_ID__FIELD.field_name: AliasUtil.to_record_id(record),
                ProcessQueueItemFields.DATA_TYPE_NAME__FIELD.field_name: AliasUtil.to_data_type_name(record)
            })
            ret_val[record] = item
        return ret_val

    def dequeue_records_for_process(self, records: Iterable[SapioRecord], wrapper: type[WrappedType],
                                    criteria: QueueItemReportCriteria = QueueItemReportCriteria()) \
            -> dict[SapioRecord, list[WrappedType]]:
        """
        Given a list of records, locate the process queue items that refer to them and that match the given search
        criteria and remove them from the queue by setting the ShowInQueue field on the process queue items to false.
        You must store and commit using the record model manager in order for these changes to take effect.

        :param records: The records to remove from the queue.
        :param wrapper: The record model wrapper for the process queue items being updated.
        :param criteria: Additional search criteria to filter the results. This function forces the show_in_queue
            parameter on the criteria to True.
        :return: A dictionary mapping each input record to the queue item records that refer to that record and were
            updated. If a record does not have any queue items that refer to it that match the given search criteria,
            then it will map to an empty list.
        """
        # Only locate queue items that are currently visible in the queue.
        criteria.shown_in_queue = True
        dequeue: dict[SapioRecord, list[WrappedType]] = self.get_queue_items_for_records(records, wrapper, criteria)
        for record, items in dequeue.items():
            for item in items:
                item.set_field_value(ProcessQueueItemFields.SHOW_IN_QUEUE__FIELD.field_name, False)
        return dequeue

    @staticmethod
    def assigned_queue_items(queue_items: Iterable[SapioRecord], assign_to: list[str]) -> None:
        """
        Given a collection of queue items, assign them to a list of usernames or group names. Only those users in the
        listed groups or those users with the matching username will be able to see these related records in the process
        queue.
        You must store and commit using the record model manager in order for these changes to take effect.

        :param queue_items: The queue items to assign.
        :param assign_to: A list of usernames and/or group names to assign these items to.
        """
        for item in queue_items:
            item.set_field_value(ProcessQueueItemFields.ASSIGNED_TO__FIELD, ",".join(assign_to))

    def get_queue_item_workflow_trackers(self, items: Iterable[SapioRecord], wrapper: type[WrappedType]) \
            -> dict[SapioRecord, list[WrappedType]]:
        """
        When a queue item is launched into a process step, a ProcessWorkflowTracking record is created by the system
        with a side link to the process queue item. This record records information about how long the record ws in
        the queue, among other details.

        Retrieve the workflow tracker records for the input queue items.

        :param items: The queue items to load the workflow trackers of.
        :param wrapper: The record model wrapper for the ProcessWorkflowTracking records.
        :return: A dictionary mapping each queue item to the workflow trackers that side link to that item. If an input
            queue item doesn't have any workflow trackers, then it will map to an empty list.
        """
        field: str = ProcessWorkflowTrackingFields.PROCESS_QUEUE_ITEM__FIELD.field_name
        self.rec_handler.rel_man.load_reverse_side_links_of_type(list(items), wrapper, field)
        return self.rec_handler.map_to_reverse_side_links(items, field, wrapper)

    @staticmethod
    def build_queue_item_report(criteria: QueueItemReportCriteria) -> CustomReportCriteria:
        """
        Construct a custom report using the provided QueueItemReportCriteria.

        :param criteria: The criteria to construct a custom report from.
        :return: A custom report that can be used to search for queue items that match the given criteria.
        """
        dt: str = ProcessQueueItemFields.DATA_TYPE_NAME
        report_builder = CustomReportBuilder(dt)
        tb = report_builder.get_term_builder()
        report_builder.add_column(SystemFields.RECORD_ID__FIELD)
        report_builder.add_column(ProcessQueueItemFields.DATA_RECORD_ID__FIELD)

        root = tb.all_records_term()

        if criteria.process_names is not None:
            term = tb.is_term(ProcessQueueItemFields.PROCESS_HEADER_NAME__FIELD, criteria.process_names)
            root = tb.and_terms(root, term)
        if criteria.not_process_names is not None:
            term = tb.not_term(ProcessQueueItemFields.PROCESS_HEADER_NAME__FIELD, criteria.not_process_names)
            root = tb.and_terms(root, term)

        if criteria.step_names is not None:
            term = tb.is_term(ProcessQueueItemFields.WORKFLOW_HEADER_NAME__FIELD, criteria.step_names)
            root = tb.and_terms(root, term)
        if criteria.not_step_names is not None:
            term = tb.not_term(ProcessQueueItemFields.WORKFLOW_HEADER_NAME__FIELD, criteria.not_step_names)
            root = tb.and_terms(root, term)

        if criteria.data_type_names is not None:
            term = tb.is_term(ProcessQueueItemFields.DATA_TYPE_NAME__FIELD, criteria.data_type_names)
            root = tb.and_terms(root, term)
        if criteria.not_data_type_names is not None:
            term = tb.not_term(ProcessQueueItemFields.DATA_TYPE_NAME__FIELD, criteria.not_data_type_names)
            root = tb.and_terms(root, term)

        if criteria.data_record_ids is not None:
            term = tb.is_term(ProcessQueueItemFields.DATA_RECORD_ID__FIELD, criteria.data_record_ids)
            root = tb.and_terms(root, term)
        if criteria.not_data_record_ids is not None:
            term = tb.not_term(ProcessQueueItemFields.DATA_RECORD_ID__FIELD, criteria.not_data_record_ids)
            root = tb.and_terms(root, term)

        if criteria.assigned_to is not None:
            term = tb.is_term(ProcessQueueItemFields.ASSIGNED_TO__FIELD, criteria.assigned_to)
            root = tb.and_terms(root, term)
        if criteria.not_assigned_to is not None:
            term = tb.not_term(ProcessQueueItemFields.ASSIGNED_TO__FIELD, criteria.not_assigned_to)
            root = tb.and_terms(root, term)

        if criteria.launched_after is not None:
            term = tb.gte_term(ProcessQueueItemFields.LAUNCHED_DATE__FIELD, criteria.launched_after)
            root = tb.and_terms(root, term)
        if criteria.launched_before is not None:
            term = tb.lte_term(ProcessQueueItemFields.LAUNCHED_DATE__FIELD, criteria.launched_before)
            root = tb.and_terms(root, term)

        if criteria.scheduled_after is not None:
            term = tb.gte_term(ProcessQueueItemFields.LAUNCHED_DATE__FIELD, criteria.scheduled_after)
            root = tb.and_terms(root, term)
        if criteria.scheduled_before is not None:
            term = tb.lte_term(ProcessQueueItemFields.LAUNCHED_DATE__FIELD, criteria.scheduled_before)
            root = tb.and_terms(root, term)

        if criteria.shown_in_queue is not None:
            term = tb.is_term(ProcessQueueItemFields.SHOW_IN_QUEUE__FIELD, criteria.shown_in_queue)
            root = tb.and_terms(root, term)
        if criteria.has_experiment is not None:
            if criteria.has_experiment:
                term = tb.not_term(ProcessQueueItemFields.EXPERIMENT__FIELD, None)
            else:
                term = tb.is_term(ProcessQueueItemFields.EXPERIMENT__FIELD, None)
            root = tb.and_terms(root, term)

        report_builder.set_root_term(root)
        return report_builder.build_report_criteria()



from sapiopylib.rest.pojo.CustomReport import ReportColumn, CustomReportCriteria, AbstractReportTerm, \
    ExplicitJoinDefinition, RelatedRecordCriteria, QueryRestriction, FieldCompareReportTerm
from sapiopylib.rest.pojo.datatype.FieldDefinition import FieldType

from sapiopycommons.customreport.column_builder import ColumnBuilder
from sapiopycommons.customreport.term_builder import TermBuilder
from sapiopycommons.general.aliases import DataTypeIdentifier, FieldIdentifier, AliasUtil, SapioRecord
from sapiopycommons.general.exceptions import SapioException


class CustomReportBuilder:
    """
    A class used for building custom reports. Look into using the TermBuilder and ColumnBuilder classes for building
    parts of a custom report.
    """
    root_data_type: DataTypeIdentifier
    data_type_name: str
    root_term: AbstractReportTerm | None
    record_criteria: RelatedRecordCriteria
    column_list: list[ReportColumn]
    join_list: list[ExplicitJoinDefinition]

    def __init__(self, root_data_type: DataTypeIdentifier):
        """
        :param root_data_type: An object that can be used to identify a data type name. Used as the root data type name
            of this search.
        """
        self.root_data_type = root_data_type
        self.data_type_name = AliasUtil.to_data_type_name(root_data_type)
        self.root_term = None
        self.record_criteria = RelatedRecordCriteria(QueryRestriction.QUERY_ALL)
        self.column_list = []
        self.join_list = []

    def get_term_builder(self) -> TermBuilder:
        """
        :return: A TermBuilder with a data type matching this report builder's root data type.
        """
        return TermBuilder(self.root_data_type)

    def has_root_term(self) -> bool:
        """
        :return: Whether this report builder has had its root term set.
        """
        return self.root_term is not None

    def set_root_term(self, term: AbstractReportTerm) -> None:
        """
        Set the root term of the report. Use the TermBuilder class to construct the report terms.

        :param term: The term to set as the root term.
        """
        self.root_term = term

    def has_columns(self) -> bool:
        """
        :return: Whether this report builder has any report columns.
        """
        return bool(self.column_list)

    def add_column(self, field: FieldIdentifier, field_type: FieldType = None,
                   *, data_type: DataTypeIdentifier | None = None) -> None:
        """
        Add a column to this report builder.

        :param field: An object that can be used to identify a data field.
        :param field_type: The field type of the provided field. This is only required if the field type cannot be
            determined from the given data type and field, which occurs when the given field is a string and the
            given data type is not a wrapped record model or record model wrapper.
        :param data_type: An object that can be used to identify a data type. If not provided, uses the root data type
            provided when this builder was initialized. You'll only want to specify this value when adding a column
            that is from a different data type than the root data type.
        """
        if data_type is None:
            data_type = self.root_data_type
        self.column_list.append(ColumnBuilder.build_column(data_type, field, field_type))

    def add_columns(self, fields: list[FieldIdentifier], *, data_type: DataTypeIdentifier | None = None) -> None:
        """
        Add columns to this report builder.

        :param fields: A list of objects that can be used to identify data fields.
        :param data_type: An object that can be used to identify a data type. If not provided, uses the root data type
            provided when this builder was initialized. You'll only want to specify this value when adding a column
            that is from a different data type than the root data type.
        """
        for field in fields:
            self.add_column(field, data_type=data_type)

    def set_query_restriction(self, base_record: SapioRecord, search_related: QueryRestriction) -> None:
        """
        Set a restriction on the report for this report builder such that the returned results must be related in
        some way to the provided base record. Without this, the report searches all records in the system that match the
        root term.

        :param base_record: The base record to run the search from.
        :param search_related: Determine the relationship of the related records that can appear in the search, be those
            children, parents, descendants, or ancestors.
        """
        if search_related == QueryRestriction.QUERY_ALL:
            raise SapioException("The search_related must be something other than QUERY_ALL when setting a query restriction.")
        self.record_criteria = RelatedRecordCriteria(search_related,
                                                     AliasUtil.to_record_id(base_record),
                                                     AliasUtil.to_data_type_name(base_record))

    def add_join(self, comparison_term: FieldCompareReportTerm, data_type: DataTypeIdentifier | None = None) -> None:
        """
        Add a join statement to this report builder.

        :param comparison_term: The field comparison term to join with.
        :param data_type: The data type name that this join is on. If not provided, then the left side data type name
            of the comparison term will be the data type that is joined against.
        """
        if data_type is None:
            data_type: str = comparison_term.left_data_type_name
        else:
            data_type: str = AliasUtil.to_data_type_name(data_type)
        self.join_list.append(ExplicitJoinDefinition(data_type, comparison_term))

    def build_report_criteria(self, page_size: int = 0, page_number: int = -1, case_sensitive: bool = False,
                              owner_restriction_set: list[str] = None) -> CustomReportCriteria:
        """
        Generate a CustomReportCriteria using the column list, root term, and root data type from this report builder.
        You can use the CustomReportManager or CustomReportUtil to run the constructed report.

        :param page_size: The page size of the custom report.
        :param page_number: The page number of the current report.
        :param case_sensitive: When searching texts, should the search be case-sensitive?
        :param owner_restriction_set: Specifies to only return records if the record is owned by this list of usernames.
        :return: A CustomReportCriteria from this report builder.
        """
        if not self.has_root_term():
            raise SapioException("Cannot build a report with no root term.")
        if not self.has_columns():
            raise SapioException("Cannot build a report with no columns.")
        return CustomReportCriteria(self.column_list, self.root_term, self.record_criteria, self.data_type_name,
                                    case_sensitive, page_size, page_number, owner_restriction_set, self.join_list)



from sapiopylib.rest.pojo.DateRange import DateRange
from sapiopylib.rest.pojo.Sort import SortDirection
from sapiopylib.rest.pojo.datatype.FieldDefinition import VeloxStringFieldDefinition, SapioStringFormat, \
    FieldValidator, VeloxAccessionFieldDefinition, VeloxBooleanFieldDefinition, VeloxDateFieldDefinition, \
    VeloxDateRangeFieldDefinition, VeloxDoubleFieldDefinition, VeloxEnumFieldDefinition, VeloxIntegerFieldDefinition, \
    VeloxLongFieldDefinition, VeloxPickListFieldDefinition, VeloxSelectionFieldDefinition, VeloxShortFieldDefinition, \
    SapioDoubleFormat, ListMode

from sapiopycommons.general.aliases import FieldIdentifier, DataTypeIdentifier, AliasUtil
from sapiopycommons.general.exceptions import SapioException


class AnyFieldInfo:
    """
    Field definition information that can apply to any created field def. This excludes various members of
    AbstractVeloxFieldDefinition such as system_field that wouldn't make sense to edit for a created field def
    used in a temp data type.
    """
    editable: bool
    required: bool
    visible: bool
    description: str | None
    sort_direction: SortDirection | None
    sort_order: int | None
    default_table_column_width: int | None

    def __init__(self, editable: bool = True, required: bool = False, visible: bool = True,
                 description: str | None = None, sort_direction: SortDirection | None = None,
                 sort_order: int | None = None, default_table_column_width: int | None = None):
        """
        :param editable: Whether this field can be edited by the user.
        :param required: Whether input is required for this field before the user can submit the dialog that it is a
            part of.
        :param visible: Whether this field is visible to the user.
        :param description: The description of this field that will appear when the user hovers the cursor over it.
        :param sort_direction: The default sort direction of this field in tables. The user may still change the column
            sorting.
        :param sort_order: The default sort order of this field in tables. The user may still change the column sorting.
        :param default_table_column_width: The width in pixels that this field's column will appear with in tables by
            default. The user may still change the column width.
        """
        self.editable = editable
        self.required = required
        self.visible = visible
        self.description = description
        self.sort_direction = sort_direction
        self.sort_order = sort_order
        self.default_table_column_width = default_table_column_width


class FieldBuilder:
    """
    A class used for building fields for temporary data types. Currently designed to only create fields which can
    be used in client callbacks that use temp data types. Some fields will not be displayed in temp data types,
    including but not limited to: action fields, child/parent/side link fields.
    """
    data_type: str

    def __init__(self, data_type: DataTypeIdentifier = "Default"):
        """
        :param data_type: The data type name that fields created from this builder will use as their data type.
        """
        self.data_type = AliasUtil.to_data_type_name(data_type)

    def accession_field(self, field_name: FieldIdentifier, sequence_key: str, prefix: str | None = None,
                        suffix: str | None = None, number_of_digits: int = 8, starting_value: int = 1,
                        link_out: dict[str, str] | None = None, abstract_info: AnyFieldInfo | None = None, *,
                        data_type_name: DataTypeIdentifier | None = None, display_name: str | None = None) \
            -> VeloxAccessionFieldDefinition:
        """
        Create an accession field definition. Accession fields are text fields which generate a unique value
        that has not been used before, incrementing from the most recently generated value. This can be used when a
        guaranteed unique ID is necessary.
    
        :param field_name: The data field name of this field. Unless a display name is also provided, this doubles as
            the display name.
        :param sequence_key: The key of the accession sequence that this field will use.
        :param prefix: The text that should appear before the numerical value.
        :param suffix: The text that should appear after the numerical value.
        :param number_of_digits: The number of digits in the numerical value.
        :param starting_value: The starting value of the numerical value.
        :param link_out: A dictionary where the keys are the display names of the links and the values are the links to
            navigate the user to if this field is clicked on. If the values contain the string "[[LINK_OUT]]" then that
            macro will be replaced with the value of the string field when it is clicked. The display name is only
            important if there is more than one link in the dictionary, in which case all available link out locations
            will display in a dialog with their display names for the user to select. If a non-empty dictionary is
            provided, this becomes a link-out field.
            If the value is not determined to have the appearance of a URL (e.g. it doesn't start with https://), then
            the system will prepend "https://<app-url>/veloxClient/" to the start of the URL. This allows you to create
            links to other locations in the system without needing to know what the app URL is. For example, if you have
            a link out string field that contains a record ID to a Sample, you could set the link value to
            "#dataType=Sample;recordId=[[LINK_OUT]];view=dataRecord" and the client will, seeing that this is not a
            normal looking URL, route the user to
            https://<app-url>/veloxClient/#dataType=Sample;recordId=[[LINK_OUT]];view=dataRecord, which is the form
            page of the Sample corresponding to the record ID recorded by the field value.
        :param abstract_info: The abstract field info for this field, such as whether it is editable or required.
        :param data_type_name: An optional override for the data type name used for this field. If not provided, then
            the data field name of the FieldBuilder is used.
        :param display_name: An optional override for the display name of this field. If not provided, then the data
            field name doubles as the display name.
        :return: An accession field definition with settings from the input criteria.
        """
        data_type_name: str = AliasUtil.to_data_type_name(data_type_name) if data_type_name else self.data_type
        field_name: str = AliasUtil.to_data_field_name(field_name)
        if abstract_info is None:
            abstract_info = AnyFieldInfo()
        if not display_name:
            display_name = field_name
        # Accession fields lock editable to false.
        abstract_info.editable = False
        link_out, link_out_url = self._convert_link_out(link_out)
        # The unique parameter has no effect, so just always set it to false.
        return VeloxAccessionFieldDefinition(data_type_name, field_name, display_name, sequence_key, prefix, suffix,
                                             number_of_digits, False, starting_value, link_out, link_out_url,
                                             **abstract_info.__dict__)

    def boolean_field(self, field_name: FieldIdentifier, default_value: bool | None = False,
                      abstract_info: AnyFieldInfo | None = None, *, data_type_name: DataTypeIdentifier | None = None,
                      display_name: str | None = None) -> VeloxBooleanFieldDefinition:
        """
        Create a boolean field definition. Boolean fields are fields which may have a value of true or false.
        They appear as a checkbox in the UI. Boolean fields may also have a value of null if the field is not required.

        :param field_name: The data field name of this field. Unless a display name is also provided, this doubles as
            the display name.
        :param default_value: The default value to display in this field before the user edits it.
        :param abstract_info: The abstract field info for this field, such as whether it is editable or required.
        :param data_type_name: An optional override for the data type name used for this field. If not provided, then
            the data field name of the FieldBuilder is used.
        :param display_name: An optional override for the display name of this field. If not provided, then the data
            field name doubles as the display name.
        :return: A boolean field definition with settings from the input criteria.
        """
        data_type_name: str = AliasUtil.to_data_type_name(data_type_name) if data_type_name else self.data_type
        field_name: str = AliasUtil.to_data_field_name(field_name)
        if abstract_info is None:
            abstract_info = AnyFieldInfo()
            # Boolean fields assume that they are required if no abstract info is provided.
            abstract_info.required = True
        if not display_name:
            display_name = field_name
        return VeloxBooleanFieldDefinition(data_type_name, field_name, display_name, default_value,
                                           **abstract_info.__dict__)

    def date_field(self, field_name: FieldIdentifier, default_value: int | None = None, date_time_format: str = "MMM dd, yyyy",
                   static_date: bool = False, abstract_info: AnyFieldInfo | None = None, *,
                   data_type_name: DataTypeIdentifier | None = None, display_name: str | None = None) \
            -> VeloxDateFieldDefinition:
        """
        Create a date field definition. Date fields store date and time information as an integer
        representing the number of milliseconds since the unix epoch. This timestamp is then displayed to users in a
        human-readable format.

        :param field_name: The data field name of this field. Unless a display name is also provided, this doubles as
            the display name.
        :param default_value: The default value to display in this field before the user edits it.
        :param date_time_format: The format that this date field should appear in. The date format is Java-style.
            See https://docs.oracle.com/en/java/javase/18/docs/api/java.base/java/text/SimpleDateFormat.html for more
            details.
        :param static_date: If true, this date displays in UTC regardless of the user's timezone. If false, this date
            displays the time in the user's timezone.
        :param abstract_info: The abstract field info for this field, such as whether it is editable or required.
        :param data_type_name: An optional override for the data type name used for this field. If not provided, then
            the data field name of the FieldBuilder is used.
        :param display_name: An optional override for the display name of this field. If not provided, then the data
            field name doubles as the display name.
        :return: A date field definition with settings from the input criteria.
        """
        data_type_name: str = AliasUtil.to_data_type_name(data_type_name) if data_type_name else self.data_type
        field_name: str = AliasUtil.to_data_field_name(field_name)
        if abstract_info is None:
            abstract_info = AnyFieldInfo()
        if not display_name:
            display_name = field_name
        return VeloxDateFieldDefinition(data_type_name, field_name, display_name, date_time_format, default_value,
                                        static_date, **abstract_info.__dict__)

    def date_range_field(self, field_name: FieldIdentifier, default_value: str | DateRange | None = None,
                         date_time_format: str = "MMM dd, yyyy", static_date: bool = False,
                         abstract_info: AnyFieldInfo | None = None, *, data_type_name: DataTypeIdentifier | None = None,
                         display_name: str | None = None) -> VeloxDateRangeFieldDefinition:
        """
        Create a date range field definition. Date range fields store two unix epoch timestamps as a string of the
        format "[start timestamp]/[end timestamp]". This string is then displayed to users in a human-readable
        format as two dates.

        See the DateRange class from sapiopylib for an easy means of converting to/from millisecond timestamps and a
        date range field's string value.

        :param field_name: The data field name of this field. Unless a display name is also provided, this doubles as
            the display name.
        :param default_value: The default value to display in this field before the user edits it.
        :param date_time_format: The format that this date field should appear in. The date format is Java-style.
            See https://docs.oracle.com/en/java/javase/18/docs/api/java.base/java/text/SimpleDateFormat.html for more
            details.
        :param static_date: If true, these dates display in UTC regardless of the user's timezone. If false, they
            display the time in the user's timezone.
        :param abstract_info: The abstract field info for this field, such as whether it is editable or required.
        :param data_type_name: An optional override for the data type name used for this field. If not provided, then
            the data field name of the FieldBuilder is used.
        :param display_name: An optional override for the display name of this field. If not provided, then the data
            field name doubles as the display name.
        :return: A date range field definition with settings from the input criteria.
        """
        data_type_name: str = AliasUtil.to_data_type_name(data_type_name) if data_type_name else self.data_type
        field_name: str = AliasUtil.to_data_field_name(field_name)
        if abstract_info is None:
            abstract_info = AnyFieldInfo()
        if not display_name:
            display_name = field_name
        if isinstance(default_value, DateRange):
            default_value = str(default_value)
        return VeloxDateRangeFieldDefinition(data_type_name, field_name, display_name, date_time_format, static_date,
                                             default_value, **abstract_info.__dict__)

    def double_field(self, field_name: FieldIdentifier, default_value: float | None = None, min_value: float = -10.**120,
                     max_value: float = 10.**120, precision: int = 1, double_format: SapioDoubleFormat | None = None,
                     abstract_info: AnyFieldInfo | None = None, *, data_type_name: DataTypeIdentifier | None = None,
                     display_name: str | None = None) -> VeloxDoubleFieldDefinition:
        """
        Create a double field definition. Double fields represent decimal numerical values. They can also
        be configured to represent currencies or percentages by changing the format parameter.

        :param field_name: The data field name of this field. Unless a display name is also provided, this doubles as
            the display name.
        :param default_value: The default value to display in this field before the user edits it.
        :param min_value: The minimum allowed value in this field.
        :param max_value: The maximum allowed value in this field.
        :param precision: The number of digits past the decimal point to display for this field.
        :param double_format: The format that this double field is displayed in. If no value is provided, the field
            display as a normal numerical value.
        :param abstract_info: The abstract field info for this field, such as whether it is editable or required.
        :param data_type_name: An optional override for the data type name used for this field. If not provided, then
            the data field name of the FieldBuilder is used.
        :param display_name: An optional override for the display name of this field. If not provided, then the data
            field name doubles as the display name.
        :return: A double field definition with settings from the input criteria.
        """
        data_type_name: str = AliasUtil.to_data_type_name(data_type_name) if data_type_name else self.data_type
        field_name: str = AliasUtil.to_data_field_name(field_name)
        if abstract_info is None:
            abstract_info = AnyFieldInfo()
        if not display_name:
            display_name = field_name
        return VeloxDoubleFieldDefinition(data_type_name, field_name, display_name, min_value, max_value, default_value,
                                          precision, double_format, **abstract_info.__dict__)

    def enum_field(self, field_name: FieldIdentifier, options: list[str], default_value: int | None = None,
                   abstract_info: AnyFieldInfo | None = None, *, data_type_name: DataTypeIdentifier | None = None,
                   display_name: str | None = None) -> VeloxEnumFieldDefinition:
        """
        Create an enum field definition. Enum fields allow for the display of a list of options as a field
        definition without the need of a backing method in the system like pick list or selection lists do. Note that
        when setting the default value or reading the return value of an enum field, the value is an integer
        representing the index of the values list, as opposed to a string for the exact value chosen.

        Note that this field is mainly here for completeness' sake. You can now use the static_values parameter of a
        selection list to achieve the same thing without needing to worry about the field value being the index of the
        options.

        :param field_name: The data field name of this field. Unless a display name is also provided, this doubles as
            the display name.
        :param options: The list of strings that the user may select from for this enum field. Note that when a client
            callback returns the value from an enum field, it will be the index of the option in the options list that
            the user chose.
        :param default_value: The default value to display in this field before the user edits it. This is the index of
            the option from the options list that you wish to appear as the default.
        :param abstract_info: The abstract field info for this field, such as whether it is editable or required.
        :param data_type_name: An optional override for the data type name used for this field. If not provided, then
            the data field name of the FieldBuilder is used.
        :param display_name: An optional override for the display name of this field. If not provided, then the data
            field name doubles as the display name.
        :return: An enum field definition with settings from the input criteria.
        """
        data_type_name: str = AliasUtil.to_data_type_name(data_type_name) if data_type_name else self.data_type
        field_name: str = AliasUtil.to_data_field_name(field_name)
        if abstract_info is None:
            abstract_info = AnyFieldInfo()
        if not display_name:
            display_name = field_name
        return VeloxEnumFieldDefinition(data_type_name, field_name, display_name, default_value, options,
                                        **abstract_info.__dict__)

    def int_field(self, field_name: FieldIdentifier, default_value: int | None = None, min_value: int = -2**31,
                  max_value: int = 2**31 - 1, unique_value: bool = False, abstract_info: AnyFieldInfo | None = None, *,
                  data_type_name: DataTypeIdentifier | None = None, display_name: str | None = None) \
            -> VeloxIntegerFieldDefinition:
        """
        Create an integer field definition. Integer fields are 32-bit whole numbers.

        :param field_name: The data field name of this field. Unless a display name is also provided, this doubles as
            the display name.
        :param default_value: The default value to display in this field before the user edits it.
        :param min_value: The minimum allowed value in this field.
        :param max_value: The maximum allowed value in this field.
        :param unique_value: Whether the value in this field must be unique across all temp records in the dialog.
        :param abstract_info: The abstract field info for this field, such as whether it is editable or required.
        :param data_type_name: An optional override for the data type name used for this field. If not provided, then
            the data field name of the FieldBuilder is used.
        :param display_name: An optional override for the display name of this field. If not provided, then the data
            field name doubles as the display name.
        :return: An integer field definition with settings from the input criteria.
        """
        data_type_name: str = AliasUtil.to_data_type_name(data_type_name) if data_type_name else self.data_type
        field_name: str = AliasUtil.to_data_field_name(field_name)
        if abstract_info is None:
            abstract_info = AnyFieldInfo()
        if not display_name:
            display_name = field_name
        return VeloxIntegerFieldDefinition(data_type_name, field_name, display_name, min_value, max_value,
                                           default_value, unique_value, **abstract_info.__dict__)

    def long_field(self, field_name: FieldIdentifier, default_value: int | None = None, min_value: int = -2**63,
                   max_value: int = 2**63 - 1, unique_value: bool = False, abstract_info: AnyFieldInfo | None = None, *,
                   data_type_name: DataTypeIdentifier | None = None, display_name: str | None = None) \
            -> VeloxLongFieldDefinition:
        """
        Create a long field definition. Long fields are 64-bit whole numbers.

        :param field_name: The data field name of this field. Unless a display name is also provided, this doubles as
            the display name.
        :param default_value: The default value to display in this field before the user edits it.
        :param min_value: The minimum allowed value in this field.
        :param max_value: The maximum allowed value in this field.
        :param unique_value: Whether the value in this field must be unique across all temp records in the dialog.
        :param abstract_info: The abstract field info for this field, such as whether it is editable or required.
        :param data_type_name: An optional override for the data type name used for this field. If not provided, then
            the data field name of the FieldBuilder is used.
        :param display_name: An optional override for the display name of this field. If not provided, then the data
            field name doubles as the display name.
        :return: A long field definition with settings from the input criteria.
        """
        data_type_name: str = AliasUtil.to_data_type_name(data_type_name) if data_type_name else self.data_type
        field_name: str = AliasUtil.to_data_field_name(field_name)
        if abstract_info is None:
            abstract_info = AnyFieldInfo()
        if not display_name:
            display_name = field_name
        return VeloxLongFieldDefinition(data_type_name, field_name, display_name, min_value, max_value, default_value,
                                        unique_value, **abstract_info.__dict__)

    def pick_list_field(self, field_name: FieldIdentifier, pick_list_name: str, default_value: str | None = None,
                        direct_edit: bool = False, abstract_info: AnyFieldInfo | None = None, *,
                        data_type_name: DataTypeIdentifier | None = None, display_name: str | None = None) \
            -> VeloxPickListFieldDefinition:
        """
        Create a pick list field definition. Pick list fields are string fields that display a drop-down list of options
        when being edited by a user. The list of options is backed by a pick list defined in the list manager sections
        of the app setup.

        Selection lists can do everything pick lists can do and more, so often it is better to use a selection list.

        :param field_name: The data field name of this field. Unless a display name is also provided, this doubles as
            the display name.
        :param pick_list_name: The name of the pick list to populate the options of this field.
        :param default_value: The default value to display in this field before the user edits it.
        :param direct_edit: Whether the user may input values not present in the list of options.
        :param abstract_info: The abstract field info for this field, such as whether it is editable or required.
        :param data_type_name: An optional override for the data type name used for this field. If not provided, then
            the data field name of the FieldBuilder is used.
        :param display_name: An optional override for the display name of this field. If not provided, then the data
            field name doubles as the display name.
        :return: A pick list field definition with settings from the input criteria.
        """
        data_type_name: str = AliasUtil.to_data_type_name(data_type_name) if data_type_name else self.data_type
        field_name: str = AliasUtil.to_data_field_name(field_name)
        if abstract_info is None:
            abstract_info = AnyFieldInfo()
        if not display_name:
            display_name = field_name
        return VeloxPickListFieldDefinition(data_type_name, field_name, display_name, pick_list_name, default_value,
                                            direct_edit, **abstract_info.__dict__)

    def selection_list_field(self, field_name: FieldIdentifier, default_value: str | None = None,
                             direct_edit: bool = False, multi_select: bool = False, unique_value: bool = False,
                             abstract_info: AnyFieldInfo | None = None, *, pick_list_name: str | None = None,
                             custom_report_name: str | None = None, plugin_name: str | None = None,
                             static_values: list[str] | None = None, user_list: bool = False,
                             user_group_list: bool = False, non_api_user_list: bool = False,
                             data_type_name: DataTypeIdentifier | None = None, display_name: str | None = None) \
            -> VeloxSelectionFieldDefinition:
        """
        Create a selection list field definition. Selection list fields are string fields that display a drop-down list
        of options when being edited by a user. The list of options can be populated from a number of locations,
        including pick lists, predefined searches (custom reports), all usernames or groups in the system, and more.

        Note that the different list types are mutually exclusive with one another. You must only provide the parameter
        necessary for a singular selection list type.

        :param field_name: The data field name of this field. Unless a display name is also provided, this doubles as
            the display name.
        :param default_value: The default value to display in this field before the user edits it.
        :param direct_edit: Whether the user may input values not present in the list of options.
        :param multi_select: Whether the user may select multiple options from the list of this field.
        :param unique_value: Whether the value in this field must be unique across all temp records in the dialog.
        :param pick_list_name: The name of the pick list to populate the options of this field.
        :param custom_report_name: The name of the custom report (predefined search) to populate the options of this
            field.
        :param plugin_name: The path to the plugin used to populate the options of this field.
        :param static_values: The list of string values used to populate the options of this field.
        :param user_list: Whether this field is populated by a list of all users in the system.
        :param user_group_list: Whether this field is populated by a list of all user groups in the system.
        :param non_api_user_list: Whether this field is populated by a list of all non-API users in the system.
        :param abstract_info: The abstract field info for this field, such as whether it is editable or required.
        :param data_type_name: An optional override for the data type name used for this field. If not provided, then
            the data field name of the FieldBuilder is used.
        :param display_name: An optional override for the display name of this field. If not provided, then the data
            field name doubles as the display name.
        :return: A selection list field definition with settings from the input criteria.
        """
        data_type_name: str = AliasUtil.to_data_type_name(data_type_name) if data_type_name else self.data_type
        field_name: str = AliasUtil.to_data_field_name(field_name)
        if abstract_info is None:
            abstract_info = AnyFieldInfo()
        if not display_name:
            display_name = field_name

        list_mode: ListMode | None = None
        if pick_list_name:
            list_mode = ListMode.LIST
        if custom_report_name:
            if list_mode:
                raise SapioException("Unable to set multiple list modes at once for a selection list.")
            list_mode = ListMode.REPORT
        if plugin_name:
            if list_mode:
                raise SapioException("Unable to set multiple list modes at once for a selection list.")
            list_mode = ListMode.PLUGIN
        if user_list:
            if list_mode:
                raise SapioException("Unable to set multiple list modes at once for a selection list.")
            list_mode = ListMode.USER
        if user_group_list:
            if list_mode:
                raise SapioException("Unable to set multiple list modes at once for a selection list.")
            list_mode = ListMode.USER_GROUP
        if non_api_user_list:
            if list_mode:
                raise SapioException("Unable to set multiple list modes at once for a selection list.")
            list_mode = ListMode.NON_API_USER
        if static_values:
            if list_mode:
                raise SapioException("Unable to set multiple list modes at once for a selection list.")
            # Static values don't have a list mode. Evaluate this last so that the multiple list modes check doesn't
            # need to be more complex.

        if not list_mode and static_values is None:
            raise SapioException("A list mode must be chosen for selection list fields.")
        return VeloxSelectionFieldDefinition(data_type_name, field_name, display_name,
                                             list_mode, unique_value, multi_select,
                                             default_value, pick_list_name, custom_report_name,
                                             plugin_name, direct_edit, static_values,
                                             **abstract_info.__dict__)

    def short_field(self, field_name: FieldIdentifier, default_value: int | None = None, min_value: int = -2**15,
                    max_value: int = 2**15 - 1, unique_value: bool = False, abstract_info: AnyFieldInfo | None = None,
                    *, data_type_name: DataTypeIdentifier | None = None, display_name: str | None = None) \
            -> VeloxShortFieldDefinition:
        """
        Create a short field definition. Short fields are 16-bit whole numbers.

        :param field_name: The data field name of this field. Unless a display name is also provided, this doubles as
            the display name.
        :param default_value: The default value to display in this field before the user edits it.
        :param min_value: The minimum allowed value in this field.
        :param max_value: The maximum allowed value in this field.
        :param unique_value: Whether the value in this field must be unique across all temp records in the dialog.
        :param abstract_info: The abstract field info for this field, such as whether it is editable or required.
        :param data_type_name: An optional override for the data type name used for this field. If not provided, then
            the data field name of the FieldBuilder is used.
        :param display_name: An optional override for the display name of this field. If not provided, then the data
            field name doubles as the display name.
        :return: A short field definition with settings from the input criteria.
        """
        data_type_name: str = AliasUtil.to_data_type_name(data_type_name) if data_type_name else self.data_type
        field_name: str = AliasUtil.to_data_field_name(field_name)
        if abstract_info is None:
            abstract_info = AnyFieldInfo()
        if not display_name:
            display_name = field_name
        return VeloxShortFieldDefinition(data_type_name, field_name, display_name, min_value, max_value, default_value,
                                         unique_value, **abstract_info.__dict__)

    def string_field(self, field_name: FieldIdentifier, default_value: str | None = None, max_length: int = 100,
                     unique_value: bool = False, html_editor: bool = False,
                     string_format: SapioStringFormat | None = None, num_lines: int = 1, auto_size: bool = False,
                     link_out: dict[str, str] | None = None, field_validator: FieldValidator | None = None,
                     abstract_info: AnyFieldInfo | None = None, *, data_type_name: DataTypeIdentifier | None = None,
                     display_name: str | None = None) -> VeloxStringFieldDefinition:
        """
        Create a string field definition. String fields represent text, and are highly customizable, allowing the
        field to be plain text or rich HTML, take up one line of space or multiple on a form, format as emails or
        phone numbers, or create links to other websites or other locations in the system.

        :param field_name: The data field name of this field. Unless a display name is also provided, this doubles as
            the display name.
        :param default_value: The default value to display in this field before the user edits it.
        :param max_length: The maximum allowed character length of this field.
        :param unique_value: Whether the value in this field must be unique across all temp records in the dialog.
        :param html_editor: Whether this field allows the user to use an HTML editor.
        :param string_format: The format that this string field is displayed in. If no value is provided, the field
            display as a normal string.
        :param num_lines: The number of lines of space that this field takes up on a form.
        :param auto_size: Whether this field should auto-size itself to fix the text when taking up space on a form.
        :param link_out: A dictionary where the keys are the display names of the links and the values are the links to
            navigate the user to if this field is clicked on. If the values contain the string "[[LINK_OUT]]" then that
            macro will be replaced with the value of the string field when it is clicked. The display name is only
            important if there is more than one link in the dictionary, in which case all available link out locations
            will display in a dialog with their display names for the user to select. If a non-empty dictionary is
            provided, this becomes a link-out field.
            If the value is not determined to have the appearance of a URL (e.g. it doesn't start with https://), then
            the system will prepend "https://<app-url>/veloxClient/" to the start of the URL. This allows you to create
            links to other locations in the system without needing to know what the app URL is. For example, if you have
            a link out string field that contains a record ID to a Sample, you could set the link value to
            "#dataType=Sample;recordId=[[LINK_OUT]];view=dataRecord" and the client will, seeing that this is not a
            normal looking URL, route the user to
            https://<app-url>/veloxClient/#dataType=Sample;recordId=[[LINK_OUT]];view=dataRecord, which is the form
            page of the Sample corresponding to the record ID recorded by the field value.
        :param field_validator: If provided, the user's input for this field must pass the regex of the given validator.
        :param abstract_info: The abstract field info for this field, such as whether it is editable or required.
        :param data_type_name: An optional override for the data type name used for this field. If not provided, then
            the data field name of the FieldBuilder is used.
        :param display_name: An optional override for the display name of this field. If not provided, then the data
            field name doubles as the display name.
        :return: A string field definition with settings from the input criteria.
        """
        data_type_name: str = AliasUtil.to_data_type_name(data_type_name) if data_type_name else self.data_type
        field_name: str = AliasUtil.to_data_field_name(field_name)
        if abstract_info is None:
            abstract_info = AnyFieldInfo()
        if not display_name:
            display_name = field_name
        link_out, link_out_url = self._convert_link_out(link_out)
        return VeloxStringFieldDefinition(data_type_name, field_name, display_name, default_value, max_length,
                                          unique_value, html_editor, string_format, num_lines, auto_size, link_out,
                                          link_out_url, field_validator, **abstract_info.__dict__)

    @staticmethod
    def _convert_link_out(link_out: dict[str, str] | None) -> tuple[bool, str | None]:
        """
        Given a dictionary of link-out URLs, convert them to the string format that the field definition expects.
        """
        if link_out:
            return True, "\t".join([display_name + "\t" + link for display_name, link in link_out.items()])
        return False, None



from collections.abc import Iterable

from sapiopylib.rest.DataMgmtService import DataMgmtServer
from sapiopylib.rest.User import SapioUser
from sapiopylib.rest.pojo.CustomReport import ReportColumn, CustomReport, CustomReportCriteria, RawReportTerm

from sapiopycommons.general.aliases import UserIdentifier, FieldValue, AliasUtil, FieldIdentifierKey


# FR-46064 - Initial port of PyWebhookUtils to sapiopycommons.
class CustomReportUtil:
    @staticmethod
    def run_system_report(context: UserIdentifier,
                          report_name: str,
                          filters: dict[FieldIdentifierKey, Iterable[FieldValue]] | None = None,
                          page_limit: int | None = None,
                          page_size: int | None = None,
                          page_number: int | None = None) -> list[dict[str, FieldValue]]:
        """
        Run a system report and return the results of that report as a list of dictionaries for the values of each
        column in each row.

        System reports are also known as predefined searches in the system and must be defined in the data designer for
        a specific data type. That is, saved searches created by users cannot be run using this function.

        :param context: The current webhook context or a user object to send requests from.
        :param report_name: The name of the system report to run.
        :param filters: If provided, filter the results of the report using the given mapping of headers to values to
            filter on. Only those headers that both the filters and the custom report share will take effect. That is,
            any filters that have a header name that isn't in the custom report will be ignored.
        :param page_limit: The maximum number of pages to query. If None, exhausts all possible pages.
        :param page_size: The size of each page of results in the search. If None, the page size is set by the server,
            which may be unlimited.
        :param page_number: The page number to start the search from, If None, starts on the first page. Note that the
            number of the first page is 0.
        :return: The results of the report listed row by row, mapping each cell to the header it is under. The header
            values in the dicts are the data field names of the columns.
            If two columns in the search have the same data field name but differing data type names, then the
            dictionary key to the value in the column will be "DataTypeName.DataFieldName". For example, if you
            had a Sample column with a data field name of Identifier and a Request column with the same data field name,
            then the dictionary keys for these columns would be Sample.Identifier and Request.Identifier respectively.
        """
        results: tuple = CustomReportUtil._exhaust_system_report(context, report_name, page_limit,
                                                                 page_size, page_number)
        columns: list[ReportColumn] = results[0]
        rows: list[list[FieldValue]] = results[1]
        return CustomReportUtil._process_results(rows, columns, filters)

    @staticmethod
    def run_custom_report(context: UserIdentifier,
                          report_criteria: CustomReportCriteria,
                          filters: dict[FieldIdentifierKey, Iterable[FieldValue]] | None = None,
                          page_limit: int | None = None,
                          page_size: int | None = None,
                          page_number: int | None = None) -> list[dict[str, FieldValue]]:
        """
        Run a custom report and return the results of that report as a list of dictionaries for the values of each
        column in each row.

        Custom reports are constructed by the caller, specifying the report terms and the columns that will be in the
        results. They are like advanced or predefined searches from the system, except they are constructed from
        within the webhook instead of from within the system.

        :param context: The current webhook context or a user object to send requests from.
        :param report_criteria: The custom report criteria to run.
        :param filters: If provided, filter the results of the report using the given mapping of headers to values to
            filter on. Only those headers that both the filters and the custom report share will take effect. That is,
            any filters that have a header name that isn't in the custom report will be ignored.
            Note that this parameter is only provided for parity with the other run report functions. If you need to
            filter the results of a search, it would likely be more beneficial to have just added a new term to the
            input report criteria that corresponds to the filter.
        :param page_limit: The maximum number of pages to query. If None, exhausts all possible pages.
        :param page_size: The size of each page of results in the search. If None, uses the value from the given report
            criteria. If not None, overwrites the value from the given report criteria.
        :param page_number: The page number to start the search from, If None, uses the value from the given report
            criteria. If not None, overwrites the value from the given report criteria. Note that the number of the
            first page is 0.
        :return: The results of the report listed row by row, mapping each cell to the header it is under. The header
            values in the dicts are the data field names of the columns.
            If two columns in the search have the same data field name but differing data type names, then the
            dictionary key to the value in the column will be "DataTypeName.DataFieldName". For example, if you
            had a Sample column with a data field name of Identifier and a Request column with the same data field name,
            then the dictionary keys for these columns would be Sample.Identifier and Request.Identifier respectively.
        """
        results: tuple = CustomReportUtil._exhaust_custom_report(context, report_criteria, page_limit,
                                                                 page_size, page_number)
        columns: list[ReportColumn] = results[0]
        rows: list[list[FieldValue]] = results[1]
        return CustomReportUtil._process_results(rows, columns, filters)

    @staticmethod
    def run_quick_report(context: UserIdentifier,
                         report_term: RawReportTerm,
                         filters: dict[FieldIdentifierKey, Iterable[FieldValue]] | None = None,
                         page_limit: int | None = None,
                         page_size: int | None = None,
                         page_number: int | None = None) -> list[dict[str, FieldValue]]:
        """
        Run a quick report and return the results of that report as a list of dictionaries for the values of each
        column in each row.

        Quick reports are helpful for cases where you need to query record field values in a more complex manner than
        the data record manager allows, but still simpler than a full-blown custom report. The columns that are returned
        in a quick search are every visible field from the data type that corresponds to the given report term. (Fields
        which are not marked as visible in the data designer will be excluded.)

        :param context: The current webhook context or a user object to send requests from.
        :param report_term: The raw report term to use for the quick report.
        :param filters: If provided, filter the results of the report using the given mapping of headers to values to
            filter on. Only those headers that both the filters and the custom report share will take effect. That is,
            any filters that have a header name that isn't in the custom report will be ignored.
        :param page_limit: The maximum number of pages to query. If None, exhausts all possible pages.
        :param page_size: The size of each page of results in the search. If None, the page size is set by the server,
            which may be unlimited.
        :param page_number: The page number to start the search from, If None, starts on the first page. Note that the
            number of the first page is 0.
        :return: The results of the report listed row by row, mapping each cell to the header it is under. The header
            values in the dicts are the data field names of the columns.
        """
        results: tuple = CustomReportUtil._exhaust_quick_report(context, report_term, page_limit,
                                                                page_size, page_number)
        columns: list[ReportColumn] = results[0]
        rows: list[list[FieldValue]] = results[1]
        return CustomReportUtil._process_results(rows, columns, filters)

    @staticmethod
    def get_system_report_criteria(context: UserIdentifier, report_name: str) -> CustomReport:
        """
        Retrieve a custom report from the system given the name of the report. This works by querying the system report
        with a page number and size of 1 to minimize the amount of data transfer needed to retrieve the report's config.

        System reports are also known as predefined searches in the system and must be defined in the data designer for
        a specific data type. That is, saved searches created by users cannot be run using this function.

        Using this, you can add to the root term of the search to then run a new search, or provide it to client
        callbacks or directives that take CustomReports.

        :param context: The current webhook context or a user object to send requests from.
        :param report_name: The name of the system report to run.
        :return: The CustomReport object for the given system report name.
        """
        user: SapioUser = AliasUtil.to_sapio_user(context)
        report_man = DataMgmtServer.get_custom_report_manager(user)
        return report_man.run_system_report_by_name(report_name, 1, 0)

    @staticmethod
    def _exhaust_system_report(context: UserIdentifier,
                               report_name: str,
                               page_limit: int | None,
                               page_size: int | None,
                               page_number: int | None) \
            -> tuple[list[ReportColumn], list[list[FieldValue]]]:
        """
        Given a system report, iterate over every page of the report and collect the results
        until there are no remaining pages.
        """
        user: SapioUser = AliasUtil.to_sapio_user(context)
        report_man = DataMgmtServer.get_custom_report_manager(user)

        # If a page size was provided but no page number was provided, then set the page number to 0,
        # as both parameters are necessary in order to get paged results.
        if page_size is not None and page_number is None:
            page_number = 0

        result = None
        has_next_page: bool = True
        rows: list[list[FieldValue]] = []
        cur_page: int = 1
        while has_next_page and (not page_limit or cur_page <= page_limit):
            result = report_man.run_system_report_by_name(report_name, page_size, page_number)
            page_size = result.page_size
            page_number = result.page_number + 1
            has_next_page = result.has_next_page
            rows.extend(result.result_table)
            cur_page += 1
        return result.column_list, rows

    @staticmethod
    def _exhaust_custom_report(context: UserIdentifier,
                               report: CustomReportCriteria,
                               page_limit: int | None,
                               page_size: int | None,
                               page_number: int | None) \
            -> tuple[list[ReportColumn], list[list[FieldValue]]]:
        """
        Given a custom report, iterate over every page of the report and collect the results
        until there are no remaining pages.
        """
        user: SapioUser = AliasUtil.to_sapio_user(context)
        report_man = DataMgmtServer.get_custom_report_manager(user)

        # If a page size was provided but no page number was provided, then set the page number to 0,
        # as both parameters are necessary in order to get paged results.
        if page_size is not None and page_number is None:
            page_number = 0

        result = None
        if page_size is not None:
            report.page_size = page_size
        if page_number is not None:
            report.page_number = page_number
        has_next_page: bool = True
        rows: list[list[FieldValue]] = []
        cur_page: int = 1
        while has_next_page and (not page_limit or cur_page <= page_limit):
            result = report_man.run_custom_report(report)
            report.page_size = result.page_size
            report.page_number = result.page_number + 1
            has_next_page = result.has_next_page
            rows.extend(result.result_table)
            cur_page += 1
        return result.column_list, rows

    @staticmethod
    def _exhaust_quick_report(context: UserIdentifier,
                              report_term: RawReportTerm,
                              page_limit: int | None,
                              page_size: int | None,
                              page_number: int | None) \
            -> tuple[list[ReportColumn], list[list[FieldValue]]]:
        """
        Given a quick report, iterate over every page of the report and collect the results
        until there are no remaining pages.
        """
        user: SapioUser = AliasUtil.to_sapio_user(context)
        report_man = DataMgmtServer.get_custom_report_manager(user)

        # If a page size was provided but no page number was provided, then set the page number to 0,
        # as both parameters are necessary in order to get paged results.
        if page_size is not None and page_number is None:
            page_number = 0

        result = None
        has_next_page: bool = True
        rows: list[list[FieldValue]] = []
        cur_page: int = 1
        while has_next_page and (not page_limit or cur_page <= page_limit):
            result = report_man.run_quick_report(report_term, page_size, page_number)
            page_size = result.page_size
            page_number = result.page_number + 1
            has_next_page = result.has_next_page
            rows.extend(result.result_table)
            cur_page += 1
        return result.column_list, rows

    @staticmethod
    def _process_results(rows: list[list[FieldValue]], columns: list[ReportColumn],
                         filters: dict[FieldIdentifierKey, Iterable[FieldValue]] | None) -> list[dict[str, FieldValue]]:
        """
        Given the results of a report as a list of row values and the report's columns, combine these lists to
        result in a singular list of dictionaries for each row in the results.

        If any filter criteria has been provided, also use that to filter the row.
        """
        # It may be the case that two columns have the same data field name but differing data type names.
        # If this occurs, then we need to be able to differentiate these columns in the resulting dictionary.
        prepend_dt: set[str] = set()
        encountered_names: list[str] = []
        for column in columns:
            field_name: str = column.data_field_name
            if field_name in encountered_names:
                prepend_dt.add(field_name)
            else:
                encountered_names.append(field_name)

        if filters:
            filters: dict[str, Iterable[FieldValue]] = AliasUtil.to_data_field_names_dict(filters)

        ret: list[dict[str, FieldValue]] = []
        for row in rows:
            row_data: dict[str, FieldValue] = {}
            filter_row: bool = False
            for value, column in zip(row, columns):
                header: str = column.data_field_name
                # If two columns share the same data field name, prepend the data type name of the column to the
                # data field name.
                if header in prepend_dt:
                    header = column.data_type_name + "." + header
                if filters is not None and header in filters and value not in filters.get(header):
                    filter_row = True
                    break
                row_data.update({header: value})
            if filter_row is False:
                ret.append(row_data)
        return ret



import re
from typing import Any, Callable, Iterable

from sapiopycommons.general.aliases import SapioRecord
from sapiopycommons.general.exceptions import SapioException
from sapiopycommons.general.time_util import TimeUtil
from sapiopycommons.recordmodel.record_handler import RecordHandler

FilterList = Iterable[int] | range | Callable[[int, dict[str, Any]], bool] | None
"""A FilterList is an object used to determine if a row in the file data should be skipped over. This can take the
form of am iterable (e.g. list, set) of its or a range where row indices in the list or range are skipped, or it can be
a callable function where rows are skipped if the function returns true. Callable function have two input parameters;
the first is the index of the row and the second is the dict for that row. If None, then the list isn't used as a
filter."""


class FileDataHandler:
    """
    A FileDataHandler takes in a list of dictionaries, presumably from a tokenized CSV or XLSX file, and allows for the
    filtering and modification of its contents. This can be used for simpler querying of values from a file, such as
    getting every value under a header in order to use it to query records in the system.

    Look into using this in combination with FileValidator to prepare files for the FileValidator and for use in
    data record fields.
    """
    file_data: list[dict[str, Any]]

    def __init__(self, file_data: list[dict[str, Any]]):
        """
        :param file_data: A list of dictionaries. Every dictionary in the list is expected to have the same keys.
            FileUtil.tokenize_csv and tokenize_xlsx can be used to convert a file into such a list.
            CustomReportUtil can also generate lists of dictionaries that match this criteria.
        """
        self.file_data = file_data

    def get_row(self, index: int) -> dict[str, Any]:
        """
        Get a particular row of data given its index.

        :param index: The index of the row to return.
        :return: The file data for the corresponding row.
        """
        return self.file_data[index]

    def get_rows(self, indices: list[int]) -> list[dict[str, Any]]:
        """
        Get a list of rows given their indices. Rows will be returned in the same order as the index list.

        :param indices: The indices of the rows to return.
        :return: The file data for the corresponding rows, in the same order as the input.
        """
        return [self.file_data[index] for index in indices]

    def has_headers(self, headers: list[str]) -> list[str]:
        """
        Determine if the file data for this handler has all the provided headers (dict keys). This assumes that every
        dict in the file data list has the same keys, so only the first row is used to verify the headers. The file data
        may have extra headers than just those in the provided list.

        :param headers: A list of headers/key names to check.
        :return: A list of all the headers that are in the provided headers list but not in the file data.
        """
        first_row: dict[str, Any] = self.get_row(0)
        missing_headers: list[str] = []
        for header in headers:
            if header not in first_row:
                missing_headers.append(header)
        return missing_headers

    def empty_cells(self, header: str,
                    *, whitelist: FilterList = None, blacklist: FilterList = None) -> list[int]:
        """
        Find the index of all rows under a particular header which are empty. A cell under a header is considered empty
        if either the value is None or the object type under the header has the __len__ special method and len(value)
        returns zero; this is used to return rows with empty strings and lists. (Not simply using bool(value) because we
        want explicit zero numerical values to be considered occupied.)

        :param header: The header of the column to check the contents of.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: The index of every row with an empty cell under the specified header.
        """
        indices: list[int] = []
        for i, row in enumerate(self.file_data):
            if self.skip_row(i, row, whitelist, blacklist):
                continue
            value: Any = row.get(header)
            if value is None or (hasattr(value, "__len__") and len(value) == 0):
                indices.append(i)
        return indices

    def occupied_cells(self, header: str,
                       *, whitelist: FilterList = None, blacklist: FilterList = None) -> list[int]:
        """
        Find the index of all rows under a particular header which are occupied. A cell under a header is considered
        occupied if the value is not None, or, should the object type under the header have the __len__ special method,
        len(value) returns zero; this is used to avoid returning rows with empty strings and lists. (Not simply using
        bool(value) because we want explicit zero numerical values to be considered occupied.)

        :param header: The header of the column to check the contents of.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: The index of every row with an occupied cell under the specified header.
        """
        indices: list[int] = []
        for i, row in enumerate(self.file_data):
            if self.skip_row(i, row, whitelist, blacklist):
                continue
            value: Any = row.get(header)
            if value is not None and (not hasattr(value, "__len__") or len(value) != 0):
                indices.append(i)
        return indices

    def find_values(self, header: str, values: Iterable[Any],
                    *, whitelist: FilterList = None, blacklist: FilterList = None) -> list[int]:
        """
        Get the index of the rows with a value under the header matching the input values.

        :param header: The header of the column to check the contents of.
        :param values: A collection of values that the cells in the column must be within.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: The index of every row with a value in the input collection under the specified header.
        """
        indices: list[int] = []
        for i, row in enumerate(self.file_data):
            if self.skip_row(i, row, whitelist, blacklist):
                continue
            if row.get(header) in values:
                indices.append(i)
        return indices

    def get_values_list(self, header: str,
                        *, whitelist: FilterList = None, blacklist: FilterList = None) -> list[Any]:
        """
        Get a list of every value under a specific header in order of appearance.

        :param header: The header of the column to check the contents of.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: A list of the values under the specified header.
        """
        values: list[Any] = []
        for i, row in enumerate(self.file_data):
            if self.skip_row(i, row, whitelist, blacklist):
                continue
            values.append(row.get(header))
        return values

    def get_values_set(self, header: str,
                       *, whitelist: FilterList = None, blacklist: FilterList = None) -> set[Any]:
        """
        Get a set of values under a specified header.

        :param header: The header of the column to check the contents of.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: A set of the values under the specified header.
        """
        values: set[Any] = set()
        for i, row in enumerate(self.file_data):
            if self.skip_row(i, row, whitelist, blacklist):
                continue
            values.add(row.get(header))
        return values

    def get_values_dict(self, header: str,
                        *, whitelist: FilterList = None, blacklist: FilterList = None) \
            -> dict[Any, list[int]]:
        """
        Get a dict of values under a specified header where each value is mapped to its row(s) of origin.

        :param header: The header of the column to check the contents of.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: A dict of the values under the specified header, mapping the cell values to the indices that they come
            from.
        """
        values: dict[Any, list[int]] = {}
        for i, row in enumerate(self.file_data):
            if self.skip_row(i, row, whitelist, blacklist):
                continue
            values.setdefault(row.get(header), []).append(i)
        return values

    def get_duplicates(self, header: str,
                       *, whitelist: FilterList = None, blacklist: FilterList = None) \
            -> dict[Any, list[int]]:
        """
        Get a dict of values under a specific header that appear more than once in that column where each value is
        mapped by its rows of origin.

        :param header: The header of the column to check the contents of.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: A dict of the values under the specified header, mapping the cell values to the indices that they come
            from.
        """
        values: dict[Any, list[int]] = self.get_values_dict(header, whitelist=whitelist, blacklist=blacklist)
        duplicates: dict[Any, list[int]] = {}
        for value, indices in values.items():
            if len(indices) > 1:
                duplicates.update({value: indices})
        return duplicates

    def get_by_function(self, func: Callable[[int, dict[str, Any]], bool],
                        *, whitelist: FilterList = None, blacklist: FilterList = None) -> list[int]:
        """
        Get the index of every row where some function returns true. This can be used for more complex validation than
        checking that a value is present or avoiding duplicates. Look at functions in this class like get_inside_range
        or get_in_List to see how this can be used. These example functions only check the contents of a single header,
        but you could create a function that checks across multiple headers as well.

        :param func: A callable function where the input is the data for a row and the output is a boolean.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: The index of every row where the provided function returns true.
        """
        indices: list[int] = []
        for i, row in enumerate(self.file_data):
            if self.skip_row(i, row, whitelist, blacklist):
                continue
            if func(i, row):
                indices.append(i)
        return indices

    def get_inside_range(self, header: str, min_val: float | int, max_val: float | int,
                         *, whitelist: FilterList = None, blacklist: FilterList = None) -> list[int]:
        """
        Get the index of every row with a value under the header inside a range defined by the min and max values.
        This range check is inclusive (i.e. min <= x <= max).

        :param header: The header of the column to check the contents of.
        :param min_val: The minimum allowed value of the cell.
        :param max_val: The maximum allowed value of the cell.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: The index of every row with a value under the specified header inside the range.
        """
        def func(index: int, row: dict[str, Any]) -> bool:
            nonlocal header, min_val, max_val
            return min_val <= row.get(header) <= max_val

        return self.get_by_function(func, whitelist=whitelist, blacklist=blacklist)

    def get_outside_range(self, header: str, min_val: float | int, max_val: float | int,
                          *, whitelist: FilterList = None, blacklist: FilterList = None) -> list[int]:
        """
        Get the index of every row with a value under the header inside a range defined by the min and max values.
        This range check is exclusive (i.e. x < min or max < x).

        :param header: The header of the column to check the contents of.
        :param min_val: The value that the cell may be lesser than.
        :param max_val: The value that the cell may be greater than.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: The index of every row with a value under the specified header outside the range.
        """
        def func(index: int, row: dict[str, Any]) -> bool:
            value = row.get(header)
            return value < min_val or value > max_val

        return self.get_by_function(func, whitelist=whitelist, blacklist=blacklist)

    def get_in_list(self, header: str, values: list[Any],
                    *, whitelist: FilterList = None, blacklist: FilterList = None) -> list[int]:
        """
        Get the index of every row with a value under the header that is within a given list.

        :param header: The header of the column to check the contents of.
        :param values: A list of values to check the header cells against.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: The index of every row with a value under the specified header inside the list.
        """
        return self.get_by_function(lambda i, row: row.get(header) in values, whitelist=whitelist, blacklist=blacklist)

    def get_not_in_list(self, header: str, values: list[Any],
                        *, whitelist: FilterList = None, blacklist: FilterList = None) -> list[int]:
        """
        Get the index of every row with a value under the header that is not within a given list.

        :param header: The header of the column to check the contents of.
        :param values: A list of values to check the header cells against.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: The index of every row with a value under the specified header not inside the range.
        """
        return self.get_by_function(lambda i, row: row.get(header) not in values, whitelist=whitelist, blacklist=blacklist)

    def get_matches(self, header: str, pattern: str | re.Pattern[str],
                    *, whitelist: FilterList = None, blacklist: FilterList = None) -> list[int]:
        """
        Get the index of every row with a value under the given header than matches a regex pattern. Unless you set up
        your regex pattern to require that the entire string must match, can return rows where only a substring of
        the cell matches.

        :param header: The header of the column to check the contents of.
        :param pattern: A regex pattern to run on the cells.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: The index of every row with a value under the specified header that matches the regex pattern.
        """
        def func(index: int, row: dict[str, Any]) -> bool:
            return len(re.findall(pattern, row.get(header))) > 0

        return self.get_by_function(func, whitelist=whitelist, blacklist=blacklist)

    def get_mismatches(self, header: str, pattern: str | re.Pattern[str],
                       *, whitelist: FilterList = None, blacklist: FilterList = None) -> list[int]:
        """
        Get the index of every row with a value under the given header than doesn't match a regex pattern.

        :param header: The header of the column to check the contents of.
        :param pattern: A regex pattern to run on the cells.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: The index of every row with a value under the specified header that doesn't match the regex pattern.
        """
        def func(index: int, row: dict[str, Any]) -> bool:
            return len(re.findall(pattern, row.get(header))) == 0

        return self.get_by_function(func, whitelist=whitelist, blacklist=blacklist)

    def set_defaults(self, header: str, value: Any,
                     *, whitelist: FilterList = None, blacklist: FilterList = None) -> None:
        """
        For all rows which are returned by the empty_cells function, set their value to some default. Useful for when
        a file may have "optional" headers and some default value is assumed if no value is given in the file.

        :param header: The header of the column to update the contents of.
        :param value: The value to set empty cells to.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        indices: list[int] = self.empty_cells(header, whitelist=whitelist, blacklist=blacklist)
        rows: list[dict[str, Any]] = self.get_rows(indices)
        for row in rows:
            row.update({header: value})

    def for_each(self, func: Callable[[int, dict[str, Any]], None],
                 *, whitelist: FilterList = None, blacklist: FilterList = None) -> None:
        """
        Run a function on rows in the file data. It is expected that the function updates values of the row, although
        this doesn't need to be the case. Look at functions in this class like update_timestamps or update_lists to see
        how this can be used.

        :param func: A callable function where the input is the index and data for a row. Has no output.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        for i, row in enumerate(self.file_data):
            if self.skip_row(i, row, whitelist, blacklist):
                continue
            func(i, row)

    def update_timestamps(self, header: str, time_format: str, timezone: str | int | None = None,
                          *, whitelist: FilterList = None, blacklist: FilterList = None) -> None:
        """
        Given a header whose contents contain integer timestamps since the epoch, convert them to human-readable
        date/time strings for a given format.

        :param header: The header of the column to update the contents of.
        :param time_format: The time format to convert the timestamp to. See TimeUtil for specifics.
        :param timezone: The timezone of the timestamp to convert from. If not specified, uses whatever timezone was
            set as the default by the TimeUtil class elsewhere in your project.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        def func(index: int, row: dict[str, Any]) -> None:
            row.update({header: TimeUtil.millis_to_format(row.get(header), time_format, timezone)})

        self.for_each(func, whitelist=whitelist, blacklist=blacklist)

    def update_dates(self, header: str, time_format: str, timezone: str | int | None = None,
                     *, whitelist: FilterList = None, blacklist: FilterList = None) -> None:
        """
        Given a header whose contents contain human-readable date/time strings for a given format, convert them to
        integer timestamps since the epoch.

        :param header: The header of the column to update the contents of.
        :param time_format: The time format to convert the timestamp from. See TimeUtil for specifics.
        :param timezone: The timezone of the time point to convert from. If not specified, uses whatever timezone was
            set as the default by the TimeUtil class elsewhere in your project.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        def func(index: int, row: dict[str, Any]) -> None:
            row.update({header: TimeUtil.format_to_millis(row.get(header), time_format, timezone)})

        self.for_each(func, whitelist=whitelist, blacklist=blacklist)

    def update_lists(self, header: str, separator: str = ",",
                     *, whitelist: FilterList = None, blacklist: FilterList = None) -> None:
        """
        Given a header whose contents contain strings representing a list of values, convert them to a list of strings
        by splitting the contents on a separator.

        :param header: The header of the column to update the contents of.
        :param separator: The separator to split the cell values on.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        def func(index: int, row: dict[str, Any]) -> None:
            row.update({header: row.get(header).split(separator)})

        self.for_each(func, whitelist=whitelist, blacklist=blacklist)

    def replace_values(self, header: str, replacements: dict[Any, Any],
                       *, whitelist: FilterList = None, blacklist: FilterList = None) -> None:
        """
        For every cell under a header, if the value in the cell is equivalent ot a key in the replacements set, then
        set that cell to the value for that key.

        :param header: The header of the column to update the contents of.
        :param replacements: A dictionary of values to replace to values to replace them with.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        def func(index: int, row: dict[str, Any]) -> None:
            value: Any = row.get(header)
            if value in replacements:
                row.update({header: replacements.get(value)})

        self.for_each(func, whitelist=whitelist, blacklist=blacklist)

    def expand_values(self, header: str, expansions: dict[str, str],
                      *, whitelist: FilterList = None, blacklist: FilterList = None) -> None:
        """
        Given a header whose contents contain strings, if the string contains any of the keys in the expansions dict,
        replace that substring of the cell with the value for that key. For example, a cell that contains "horse" with
        a dict {"r": "u"} would result in a cell being updated to read "house".

        Expansions are recursive; e.g. "horse" with the dict {"r": "u", "u": "r"} would replace the "r" with "u"
        would then replace that "u" with "r", resulting in no change.

        :param header: The header of the column to update the contents of.
        :param expansions: A dictionary of values to replace to values to replace them with.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        """
        def func(index: int, row: dict[str, Any]) -> None:
            value: str = row.get(header)
            for key in expansions:
                if key in value:
                    value = value.replace(key, expansions.get(key))
                    row.update({header: value})

        self.for_each(func, whitelist=whitelist, blacklist=blacklist)

    def match_records(self, header: str, field: str, records: list[SapioRecord]) -> dict[SapioRecord, dict[str, Any]]:
        """
        Match a list of records to rows in the file given a header and field to match on.

        The expectation is that no two records have the same value for the given field and that no two rows in the file
        map to the same record. It is allowable that a record may have no matching row in the file.

        :param header: The header of the column to check against.
        :param field: The data field name on the records to check against.
        :param records: The records to map to the file rows.
        :return: A dict of record to the row where that record's field value matches the row's value under the header.
        """
        mapped_records: dict[Any, SapioRecord] = RecordHandler.map_by_unique_field(records, field)
        matches: dict[SapioRecord, dict[str, Any]] = {}
        for row in self.file_data:
            value: Any = row.get(header)
            record: SapioRecord = mapped_records.get(value)
            if record:
                if record in matches:
                    raise SapioException(f"The header {header} value {value} matches with multiple records in the "
                                         f"given list of records.")
                matches.update({record: row})
        return matches

    def update_records(self, match_header: str, match_field: str, records: list[SapioRecord],
                       header_to_fields: dict[str, str]) -> None:
        """
        Match a list of records to rows in the file given a header and field ot match on, then update the fields of
        those records using values from the column of the matching row.

        The expectation is that no two records have the same value for the given field and that no two rows in the file
        map to the same record. It is allowable that a record may have no matching row in the file.

        :param match_header: The header of the column to match against.
        :param match_field: The data field name on the records to match against.
        :param records: The records to update.
        :param header_to_fields: A dict of file header to record field name to.
        """
        records_to_row: dict[SapioRecord, dict[str, Any]] = self.match_records(match_header, match_field, records)
        for record, row in records_to_row.items():
            for header, field in header_to_fields.items():
                record.set_field_value(field, row.get(header))

    def find_missing_values(self, header: str, field: str, records: list[SapioRecord]) -> list[int]:
        """
        Given a header and a list of records with a field to check, return every value that is present in the file
        but is not present in the record fields.

        :param header: The header of the column to check against.
        :param field: The data field name on the records to check against.
        :param records: The records to check if their values exist in the file.
        :return: A list of row indices for rows that contain a value that isn't present in the given records.
        """
        missing: list[int] = []
        values: list[Any] = [x.get_field_value(field) for x in records]
        for i, row in enumerate(self.file_data):
            if row.get(header) not in values:
                missing.append(i)
        return missing

    def get_differences(self, diff_check: list[dict[str, Any]], headers: list[str]) -> dict[str, list[int]]:
        """
        Given a list of dictionaries and a list of headers to check, return an index of every row where the values
        in the file and the values in the given list differ. The given list should be the same size as the file. Rows
        in the file are compared against the matching index in the diff check list.

        :param diff_check: A list of dictionaries similar to the list used to initialize this FileDataHandler. The
            number of elements in this list should be equivalent to the number of elements in hte initializing list.
            This list's dicts do not necessarily need to have all the same headers as the initializing dicts, but both
            should have the headers that are in the headers parameter.
        :param headers: A list of the specific headers that should be difference checked.
        :return: A dictionary of headers to a list of indices for the rows that differed between the file and input.
        """
        differences: dict[str, list[int]] = {}
        for i, (row, other_row) in enumerate(zip(self.file_data, diff_check, strict=True)):
            for header in headers:
                if row.get(header) != other_row.get(header):
                    differences.setdefault(header, []).append(i)
        return differences

    def get_differences_for_index(self, index: int, diff_check: dict[str, Any], headers: list[str]) -> list[str]:
        """
        Given a row index, a dictionary , and a list of headers to check, return the headers where the values
        in the row at the given index and the values in the given dict differ.

        :param index: The index of the row to check against the given dict.
        :param diff_check: A dictionary similar to the dictionaries in the list used to initialize this FileDataHandler.
            This dict does not necessarily need to have all the same headers as the initializing dicts, but both
            should have the headers that are in the headers parameter.
        :param headers: A list of the specific headers that should be difference checked.
        :return: A list of headers that differed between the file and input.
        """
        differences: list[str] = []
        row: dict[str, Any] = self.get_row(index)
        for header in headers:
            if row.get(header) != diff_check.get(header):
                differences.append(header)
        return differences

    def get_differences_for_record(self, row: dict[str, Any] | int, record: SapioRecord,
                                   field_mappings: dict[str, str]) -> list[str]:
        """
        Given a row, a record, and a dictionary mapping headers in the row to fields on the record, return a list of
        all file headers where the value in the file differs from the value in the record.

        :param row: A row from the file, either as a dictionary or an integer for the row's index.
        :param record: A record to compare against.
        :param field_mappings: A dictionary that maps header names in the file to field names on the record.
        :return: A list of file headers where the file and record have different values.
        """
        if isinstance(row, int):
            row: dict[str, Any] = self.get_row(row)

        differences: list[str] = []
        for header, field in field_mappings.items():
            if row.get(header) != record.get_field_value(field):
                differences.append(header)
        return differences

    @staticmethod
    def skip_row(i: int, row: dict[str, Any], whitelist: FilterList, blacklist: FilterList) -> bool:
        """
        Determine whether a row should be skipped given whitelist and blacklist filters.

        :param i: The index of the row.
        :param row: The file data of the row.
        :param whitelist: If a row doesn't match the whitelist, it will be skipped over. See the FilterList alias
            description for the forms that a whitelist can take.
        :param blacklist: If a row matches the blacklist, it will be skipped over. See the FilterList alias
            description for the forms that a blacklist can take.
        :return: Whether the row should be skipped.
        """
        if whitelist is not None:
            if isinstance(whitelist, (list, range)):
                return i not in whitelist
            return whitelist(i, row)
        if blacklist is not None:
            if isinstance(blacklist, (list, range)):
                return i in blacklist
            return blacklist(i, row)
        return False



from __future__ import annotations

import time
from collections.abc import Mapping, Iterable
from typing import TypeAlias
from weakref import WeakValueDictionary

from sapiopylib.rest.DataMgmtService import DataMgmtServer
from sapiopylib.rest.ELNService import ElnManager
from sapiopylib.rest.User import SapioUser
from sapiopylib.rest.pojo.DataRecord import DataRecord
from sapiopylib.rest.pojo.eln.ElnExperiment import ElnExperiment, TemplateExperimentQueryPojo, ElnTemplate, \
    InitializeNotebookExperimentPojo, ElnExperimentUpdateCriteria
from sapiopylib.rest.pojo.eln.ExperimentEntry import ExperimentEntry
from sapiopylib.rest.pojo.eln.ExperimentEntryCriteria import AbstractElnEntryUpdateCriteria
from sapiopylib.rest.pojo.eln.SapioELNEnums import ExperimentEntryStatus, ElnExperimentStatus, ElnEntryType, \
    ElnBaseDataType
from sapiopylib.rest.pojo.webhook.WebhookContext import SapioWebhookContext
from sapiopylib.rest.pojo.webhook.WebhookDirective import ElnExperimentDirective
from sapiopylib.rest.pojo.webhook.WebhookResult import SapioWebhookResult
from sapiopylib.rest.utils.Protocols import ElnEntryStep, ElnExperimentProtocol
from sapiopylib.rest.utils.recordmodel.PyRecordModel import PyRecordModel
from sapiopylib.rest.utils.recordmodel.RecordModelManager import RecordModelInstanceManager, RecordModelManager
from sapiopylib.rest.utils.recordmodel.RecordModelWrapper import WrappedType
from sapiopylib.rest.utils.recordmodel.properties import Child

from sapiopycommons.eln.experiment_report_util import ExperimentReportUtil
from sapiopycommons.general.aliases import AliasUtil, SapioRecord, ExperimentIdentifier, UserIdentifier, \
    DataTypeIdentifier, RecordModel
from sapiopycommons.general.exceptions import SapioException

Step: TypeAlias = str | ElnEntryStep
"""An object representing an identifier to an ElnEntryStep. May be either the name of the step or the ElnEntryStep
itself."""


# FR-46064 - Initial port of PyWebhookUtils to sapiopycommons.
class ExperimentHandler:
    user: SapioUser
    context: SapioWebhookContext | None
    """The context that this handler is working from."""

    # Basic experiment info from the context.
    __eln_exp: ElnExperiment
    """The ELN experiment from the context."""
    __protocol: ElnExperimentProtocol
    """The ELN experiment as a protocol."""
    __exp_id: int
    """The ID of this experiment's notebook. Used for making update webservice calls."""

    # Managers.
    __eln_man: ElnManager
    """The ELN manager. Used for updating the experiment and its steps."""
    __inst_man: RecordModelInstanceManager
    """The record model instance manager. Used for wrapping the data records of a step as record models."""

    # Only a fraction of the information about the current experiment exists in the context. Much information requires
    # additional queries to obtain, but may also be repeatedly accessed. In such cases, cache the information after it
    # has been requested so that the user doesn't need to worry about caching it themselves.
    # CR-46341: Replace class variables with instance variables.
    __exp_record: DataRecord | None
    """The data record for this experiment. Only cached when first accessed."""
    __exp_template: ElnTemplate | None
    """The template for this experiment. Only cached when first accessed."""
    __exp_options: dict[str, str]
    """Experiment options for this experiment. Only cached when first accessed."""

    __queried_all_steps: bool
    """Whether this ExperimentHandler has queried the system for all steps in the experiment."""
    __steps: dict[str, ElnEntryStep]
    """Steps from this experiment. All steps are cached the first time any individual step is accessed."""
    __step_options: dict[int, dict[str, str]]
    """Entry options for each step in this experiment. All entry options are cached the first time any individual step's
    options are queried. The cache is updated whenever the entry options for a step are changed by this handler."""

    # Constants
    __ENTRY_COMPLETE_STATUSES = [ExperimentEntryStatus.Completed, ExperimentEntryStatus.CompletedApproved]
    """The set of statuses that an ELN entry could have and be considered completed/submitted."""
    __ENTRY_LOCKED_STATUSES = [ExperimentEntryStatus.Completed, ExperimentEntryStatus.CompletedApproved,
                               ExperimentEntryStatus.Disabled, ExperimentEntryStatus.LockedAwaitingApproval,
                               ExperimentEntryStatus.LockedRejected]
    """The set of statuses that an ELN entry could have and be considered locked."""
    __EXPERIMENT_COMPLETE_STATUSES = [ElnExperimentStatus.Completed, ElnExperimentStatus.CompletedApproved]
    """The set of statuses that an ELN experiment could have and be considered completed."""
    __EXPERIMENT_LOCKED_STATUSES = [ElnExperimentStatus.Completed, ElnExperimentStatus.CompletedApproved,
                                    ElnExperimentStatus.LockedRejected, ElnExperimentStatus.LockedAwaitingApproval,
                                    ElnExperimentStatus.Canceled]
    """The set of statuses that an ELN experiment could have and be considered locked."""

    __instances: WeakValueDictionary[str, ExperimentHandler] = WeakValueDictionary()
    __initialized: bool

    def __new__(cls, context: UserIdentifier, experiment: ExperimentIdentifier | SapioRecord | None = None):
        """
        :param context: The current webhook context or a user object to send requests from.
        :param experiment: If an experiment is provided that is separate from the experiment that is in the context,
            that experiment will be used by this ExperimentHandler instead. An experiment can be provided in various
            forms, including an ElnExperiment, ElnExperimentProtocol, an experiment record, or a notebook experiment ID.
        """
        param_results = cls.__parse_params(context, experiment)
        user = param_results[0]
        experiment = param_results[2]
        key = f"{user.__hash__()}:{experiment.notebook_experiment_id}"
        obj = cls.__instances.get(key)
        if not obj:
            obj = object.__new__(cls)
            obj.__initialized = False
            cls.__instances[key] = obj
        return obj

    def __init__(self, context: UserIdentifier, experiment: ExperimentIdentifier | SapioRecord | None = None):
        """
        Initialization will throw an exception if there is no ELN Experiment in the provided context and no experiment
        is provided.

        :param context: The current webhook context or a user object to send requests from.
        :param experiment: If an experiment is provided that is separate from the experiment that is in the context,
            that experiment will be used by this ExperimentHandler instead. An experiment can be provided in various
            forms, including an ElnExperiment, ElnExperimentProtocol, an experiment record, or a notebook experiment ID.
        """
        param_results = self.__parse_params(context, experiment)
        self.user = param_results[0]
        self.context = param_results[1]
        experiment = param_results[2]

        # Get the basic information about this experiment that already exists in the context and is often used.
        self.__eln_exp = experiment
        self.__protocol = ElnExperimentProtocol(experiment, self.user)
        self.__exp_id = self.__protocol.get_id()

        # Grab various managers that may be used.
        self.__eln_man = DataMgmtServer.get_eln_manager(self.user)
        self.__inst_man = RecordModelManager(self.user).instance_manager

        # Create empty caches to fill when necessary.
        self.__steps = {}
        self.__step_options = {}
        # CR-46330: Cache any experiment entry information that might already exist in the context.
        self.__queried_all_steps = False
        # We can only trust the entries in the context if the experiment that this handler is for is the same as the
        # one from the context.
        if self.context is not None and self.context.eln_experiment == experiment:
            if self.context.experiment_entry is not None:
                self.__steps.update({self.context.active_step.get_name(): self.context.active_step})
            if self.context.experiment_entry_list is not None:
                for entry in self.context.experiment_entry_list:
                    self.__steps.update({entry.entry_name: ElnEntryStep(self.__protocol, entry)})

    @staticmethod
    def __parse_params(context: UserIdentifier, experiment: ExperimentIdentifier | SapioRecord | None = None) \
            -> tuple[SapioUser, SapioWebhookContext | None, ElnExperiment]:
        if isinstance(context, SapioWebhookContext):
            user = context.user
            context = context
        else:
            user = context
            context = None
        if context is not None and context.eln_experiment is not None and experiment is None:
            experiment = context.eln_experiment
        # FR-46495 - Allow the init function of ExperimentHandler to take in an ElnExperiment that is separate from the
        # context.
        # CR-37038 - Allow other experiment object types to be provided. Convert them all down to ElnExperiment.
        if (context is None or context.eln_experiment is None) and experiment is not None:
            eln_manager = DataMgmtServer.get_eln_manager(user)
            # If this object is already an ElnExperiment, do nothing.
            if isinstance(experiment, ElnExperiment):
                pass
            # If this object is an ElnExperimentProtocol, then we can get the ElnExperiment from the object.
            elif isinstance(experiment, ElnExperimentProtocol):
                experiment: ElnExperiment = experiment.eln_experiment
            # If this object is an integer, assume it is a notebook ID that we can query the system with.
            elif isinstance(experiment, int):
                notebook_id: int = experiment
                experiment: ElnExperiment = eln_manager.get_eln_experiment_by_id(notebook_id)
                if not experiment:
                    raise SapioException(f"No experiment with notebook ID {notebook_id} located in the system.")
            # If this object is a record, assume it is an experiment record that we can query the system with.
            else:
                record_id: int = AliasUtil.to_record_ids([experiment])[0]
                experiment: ElnExperiment = eln_manager.get_eln_experiment_by_record_id(record_id)
                if not experiment:
                    raise SapioException(f"No experiment with record ID {record_id} located in the system.")
        if experiment is None:
            raise SapioException("Cannot initialize ExperimentHandler. No ELN Experiment found in the provided parameters.")

        return user, context, experiment

    # FR-46495: Split the creation of the experiment in launch_experiment into a create_experiment function.
    @staticmethod
    def create_experiment(context: SapioWebhookContext,
                          template_name: str,
                          experiment_name: str | None = None,
                          parent_record: SapioRecord | None = None, *,
                          template_version: int | None = None, active_templates_only: bool = True) -> ElnExperiment:
        """
        Create an ElnExperiment from the given template name.

        Makes a webservice request to query for all the templates matching the provided criteria. Note that if multiple
        templates match the same criteria, the first template that is encountered in the query is used. Throws an
        exception if no template is found. Also makes a webservice request to create the experiment.

        :param context: The current webhook context.
        :param template_name: The name of the template to create the experiment from.
        :param experiment_name: The name to give to the experiment after it is created. If not provided, defaults to the
            display name of the template.
        :param parent_record: The parent record to attach this experiment under. This record must be an eligible
            parent type to ELNExperiment. If not provided, the experiment is stored in the aether.
        :param template_version: The version number of the template to use. If not provided, the latest version of the
            template is used. NOTICE: Template version numbers aren't necessarily the same between environments, so
            be careful with using the same webhook across multiple environments if you are searching for a specific
            version number.
        :param active_templates_only: Whether only active templates should be queried for.
        :return: The newly created experiment.
        """
        template_query = TemplateExperimentQueryPojo(latest_version_only=(template_version is None),
                                                     active_templates_only=active_templates_only)
        templates: list[ElnTemplate] = context.eln_manager.get_template_experiment_list(template_query)
        launch_template: ElnTemplate | None = None
        for template in templates:
            if template.template_name != template_name:
                continue
            if template_version is not None and template.template_version != template_version:
                continue
            launch_template = template
            break
        if launch_template is None:
            raise SapioException(f"No template with the name \"{template_name}\"" +
                                 ("" if template_version is None else f" and the version {template_version}") +
                                 f" found.")

        if experiment_name is None:
            experiment_name: str = launch_template.display_name
        if parent_record is not None:
            parent_record: DataRecord = AliasUtil.to_data_record(parent_record)
        notebook_init = InitializeNotebookExperimentPojo(experiment_name, launch_template.template_id, parent_record)
        return context.eln_manager.create_notebook_experiment(notebook_init)

    @staticmethod
    def launch_experiment(context: SapioWebhookContext,
                          template_name: str,
                          experiment_name: str | None = None,
                          parent_record: SapioRecord | None = None, *,
                          template_version: int | None = None, active_templates_only: bool = True) -> SapioWebhookResult:
        """
        Create a SapioWebhookResult that, when returned by a webhook handler, sends the user to a new experiment of the
        input template name.

        Makes a webservice request to query for all the templates matching the provided criteria. Note that if multiple
        templates match the same criteria, the first template that is encountered in the query is used. Throws an
        exception if no template is found. Also makes a webservice request to create the experiment.

        :param context: The current webhook context.
        :param template_name: The name of the template to create the experiment from.
        :param experiment_name: The name to give to the experiment after it is created. If not provided, defaults to the
            display name of the template.
        :param parent_record: The parent record to attach this experiment under. This record must be an eligible
            parent type to ELNExperiment. If not provided, the experiment is stored in the aether.
        :param template_version: The version number of the template to use. If not provided, the latest version of the
            template is used. NOTICE: Template version numbers aren't necessarily the same between environments, so
            be careful with using the same webhook across multiple environments if you are searching for a specific
            version number.
        :param active_templates_only: Whether only active templates should be queried for.
        :return: A SapioWebhookResult that directs the user to the newly created experiment.
        """
        experiment = ExperimentHandler.create_experiment(context, template_name, experiment_name, parent_record,
                                                         template_version=template_version,
                                                         active_templates_only=active_templates_only)
        return SapioWebhookResult(True, directive=ElnExperimentDirective(experiment.notebook_experiment_id))

    def get_experiment_template(self, exception_on_none: bool = True) -> ElnTemplate | None:
        """
        Query for the experiment template for the current experiment. The returned template is cached by the
        ExperimentHandler.

        :param exception_on_none: If false, returns None if there is no experiment template. If true, raises an exception
            when the experiment template doesn't exist.
        :return: This experiment's template. None if it has no template.
        """
        template_id: int | None = self.__eln_exp.template_id
        if template_id is None:
            self.__exp_template = None
            if exception_on_none:
                raise SapioException(f"Experiment with ID {self.__exp_id} has no template ID.")
            return None

        if not hasattr(self, "_ExperimentHandler__exp_template"):
            # PR-46504: Allow inactive and non-latest version templates to be queried.
            query = TemplateExperimentQueryPojo(template_id_white_list=[template_id],
                                                active_templates_only=False,
                                                latest_version_only=False)
            templates: list[ElnTemplate] = self.__eln_man.get_template_experiment_list(query)
            # PR-46504: Set the exp_template to None if there are no results.
            self.__exp_template = templates[0] if templates else None
        if self.__exp_template is None and exception_on_none:
            raise SapioException(f"Experiment template not found for experiment with ID {self.__exp_id}.")
        return self.__exp_template

    # CR-46104: Change get_template_name to behave like NotebookProtocolImpl.getTemplateName (i.e. first see if the
    # experiment template exists, and if not, see if the experiment record exists, instead of only checking the
    # template).
    def get_template_name(self, exception_on_none: bool = True) -> str | None:
        """
        Get the template name for the current experiment.

        The template name is determined by either the experiment template or the experiment record, whichever is
        already cached. If neither are cached, queries for the experiment template. If no experiment template exists,
        queries for the experiment record.

        :param exception_on_none: If false, returns None if there is no template name. If true, raises an exception
            when the template name doesn't exist.
        :return: The template name of the current experiment. None if it has no template name.
        """
        if not hasattr(self, "_ExperimentHandler__exp_template"):
            self.get_experiment_template(False)
        if self.__exp_template is None and not hasattr(self, "_ExperimentHandler__exp_record"):
            self.get_experiment_record(False)

        name: str | None = None
        if self.__exp_template is not None:
            name = self.__exp_template.template_name
        elif self.__exp_record is not None:
            name = self.__exp_record.get_field_value("TemplateExperimentName")
        if name is None and exception_on_none:
            raise SapioException(f"Template name not found for experiment with ID {self.__exp_id}.")
        return name

    def get_experiment_record(self, exception_on_none: bool = True) -> DataRecord | None:
        """
        Query for the data record of this experiment. The returned record is cached by the ExperimentHandler.

        :param exception_on_none: If false, returns None if there is no experiment record. If true, raises an exception
            when the experiment record doesn't exist.
        :return: The data record for this experiment. None if it has no record.
        """
        if not hasattr(self, "_ExperimentHandler__exp_record"):
            self.__exp_record = self.__protocol.get_record()
        if self.__exp_record is None and exception_on_none:
            raise SapioException(f"Experiment record not found for experiment with ID {self.__exp_id}.")
        return self.__exp_record

    def get_experiment_model(self, wrapper_type: type[WrappedType]) -> WrappedType:
        """
        Query for the data record of this experiment and wrap it as a record model with the given wrapper.
        The returned record is cached by the ExperimentHandler.

        :param wrapper_type: The record model wrapper to use.
        :return: The record model for this experiment.
        """
        return self.__inst_man.add_existing_record_of_type(self.get_experiment_record(), wrapper_type)

    def update_experiment(self,
                          experiment_name: str | None = None,
                          experiment_status: ElnExperimentStatus | None = None,
                          experiment_option_map: dict[str, str] | None = None) -> None:
        """
        Make a webservice call to update the experiment for this ExperimentHandler.  If any parameter is not provided,
        then no change is made to it.

        :param experiment_name: The new name of the experiment.
        :param experiment_status: The new status of this experiment.
        :param experiment_option_map:
            The new map of options for this experiment. Completely overwrites the existing options map.
            Any changes to the experiment options will update this ExperimentHandler's cache of the experiment options.
            If you wish to add options to the existing map of options that an experiment has, use the
            add_experiment_options method.
        """
        criteria = ElnExperimentUpdateCriteria()
        criteria.new_experiment_name = experiment_name
        criteria.new_experiment_status = experiment_status
        criteria.experiment_option_map = experiment_option_map
        self.__eln_man.update_notebook_experiment(self.__exp_id, criteria)

        if experiment_name is not None:
            self.__eln_exp.notebook_experiment_name = experiment_name
        if experiment_status is not None:
            self.__eln_exp.notebook_experiment_status = experiment_status
        if experiment_option_map is not None:
            self.__exp_options = experiment_option_map

    def get_experiment_option(self, option: str) -> str:
        """
        Get the value of a specific experiment option.

        Getting the experiment options requires a webservice query, which is made the first time any experiment option
        method is called by this ExperimentHandler. The experiment options are cached so that subsequent calls of this
        method don't make a webservice call.

        :param option: The experiment option to search for.
        :return: The value of the input experiment options.
        """
        return self.get_experiment_options().get(option)

    def get_experiment_options(self) -> dict[str, str]:
        """
        Get the entire map of options for this experiment.

        Getting the experiment options requires a webservice query, which is made the first time any experiment option
        method is called by this ExperimentHandler. The experiment options are cached so that subsequent calls of this
        method don't make a webservice call.

        :return: The map of options for this experiment.
        """
        return self.__get_experiment_options()

    def add_experiment_options(self, mapping: Mapping[str, str]) -> None:
        """
        Add to the existing map of options for this experiment. Makes a webservice call to update the experiment. Also
        updates the cache of the experiment options.

        Getting the experiment options requires a webservice query, which is made the first time any experiment option
        method is called by this ExperimentHandler. The experiment options are cached so that subsequent calls of this
        method don't make a webservice call.

        :param mapping: The new options and values to add to the existing experiment options, provided as some Mapping
            (e.g. a Dict). If an option key already exists and is provided in the mapping, overwrites the existing value
            for that key.
        """
        options: dict[str, str] = self.get_experiment_options()
        options.update(mapping)
        self.update_experiment(experiment_option_map=options)

    def experiment_is_complete(self) -> bool:
        """
        Determine if the experiment has been completed.

        :return: True if the experiment's status is Completed or CompletedApproved. False otherwise.
        """
        return self.__eln_exp.notebook_experiment_status in self.__EXPERIMENT_COMPLETE_STATUSES

    def experiment_is_canceled(self) -> bool:
        """
        Determine if the experiment has been canceled.

        :return: True if the experiment's status is Canceled. False otherwise.
        """
        return self.__eln_exp.notebook_experiment_status == ElnExperimentStatus.Canceled

    def experiment_is_locked(self) -> bool:
        """
        Determine if the experiment has been locked in any way.

        :return: True if the experiment's status is Completed, CompletedApproved, Canceled, LockedAwaitingApproval,
            or LockedRejected. False otherwise.
        """
        return self.__eln_exp.notebook_experiment_status in self.__EXPERIMENT_LOCKED_STATUSES

    def complete_experiment(self) -> None:
        """
        Set the experiment's status to Completed. Makes a webservice call to update the experiment. Checks if the
        experiment is already completed, and does nothing if so.
        """
        if not self.experiment_is_complete():
            self.__protocol.complete_protocol()
            self.__eln_exp.notebook_experiment_status = ElnExperimentStatus.Completed

    def cancel_experiment(self) -> None:
        """
        Set the experiment's status to Canceled. Makes a webservice call to update the experiment. Checks if the
        experiment is already canceled, and does nothing if so.

        NOTE: This will not run the usual logic around canceling an experiment that you'd see if canceling the
        experiment using the "Cancel Experiment" toolbar button, such as moving in process samples back to the queue,
        as those changes are tied to the button instead of being on the experiment status change.
        """
        if not self.experiment_is_canceled():
            self.__protocol.cancel_protocol()
            self.__eln_exp.notebook_experiment_status = ElnExperimentStatus.Canceled

    def step_exists(self, step_name: str) -> bool:
        """
        Determine if a step by the given name exists in the experiment.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step_name: The name of the step to search for.
        :return: True if the step exists, false otherwise.
        """
        return self.get_step(step_name, False) is not None

    def steps_exist(self, step_names: Iterable[str]) -> bool:
        """
        Determine if all the steps by the given names exist in the experiment.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step_names: The names of the steps to search for.
        :return: True if every step exists, false if at least one does not exist.
        """
        return all([x is not None for x in self.get_steps(step_names, False)])

    def get_step(self, step_name: str, exception_on_none: bool = True) -> ElnEntryStep | None:
        """
        Get the step of the given name from the experiment.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step_name: The name for the step to return.
        :param exception_on_none: If false, returns None if the entry can't be found. If true, raises an exception
            when the named entry doesn't exist in the experiment.
        :return: An ElnEntrySteps matching the provided name. If there is no match and no exception is to be thrown,
            returns None.
        """
        return self.get_steps([step_name], exception_on_none)[0]

    def get_steps(self, step_names: Iterable[str], exception_on_none: bool = True) -> list[ElnEntryStep | None]:
        """
        Get a list of steps of the given names from the experiment, sorted in the same order as the names are provided.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step_names: A list of names for the entries to return and the order to return them in.
        :param exception_on_none: If false, returns None for entries that can't be found. If true, raises an exception
            when the named entry doesn't exist in the experiment.
        :return: A list of ElnEntrySteps matching the provided names in the order they were provided in. If there is no
            match for a given step and no exception is to be thrown, returns None for that step.
        """
        ret_list: list[ElnEntryStep | None] = []
        for name in step_names:
            # If we haven't queried the system for all steps in the experiment yet, then the reason that a step is
            # missing may be because it wasn't in the webhook context. Therefore, query all steps and then check
            # if the step name is still missing from the experiment before potentially throwing an exception.
            if self.__queried_all_steps is False and name not in self.__steps:
                self.__queried_all_steps = True
                self.__steps.update({step.get_name(): step for step in self.__protocol.get_sorted_step_list()})

            step: ElnEntryStep = self.__steps.get(name)
            if step is None and exception_on_none is True:
                raise SapioException(f"ElnEntryStep of name \"{name}\" not found in experiment with ID {self.__exp_id}.")
            ret_list.append(step)
        return ret_list

    def get_all_steps(self, data_type: DataTypeIdentifier | None = None) -> list[ElnEntryStep]:
        """
        Get a list of every entry in the experiment. Optionally filter the returned entries by a data type.

        Makes a webservice call to retrieve every entry in the experiment if they were not already previously cached.

        :param data_type: A data type used to filter the returned entries. If None is given, returns all entries. If
            a data type name or wrapper is given, only returns entries that match that data type name or wrapper.
        :return: Every entry in the experiment in order of appearance that match the provided data type, if any.
        """
        if self.__queried_all_steps is False:
            self.__queried_all_steps = True
            self.__steps.update({step.get_name(): step for step in self.__protocol.get_sorted_step_list()})
        all_steps: list[ElnEntryStep] = self.__protocol.get_sorted_step_list()
        if data_type is None:
            return all_steps
        data_type: str = AliasUtil.to_data_type_name(data_type)
        return [x for x in all_steps if data_type in x.get_data_type_names()]

    def get_step_by_option(self, key: str, value: str | None = None) -> ElnEntryStep:
        """
        Retrieve the step in this experiment that contains an entry option with the provided key and value.
        Throws an exception if no entries or multiple entries in the experiment match.

        :param key: The key of the entry option to match on.
        :param value: The value of the entry option to match on. If not provided, then only matches on key.
        :return: The entry in this experiment that matches the provided entry option key and value.
        """
        steps: list[ElnEntryStep] = self.get_steps_by_option(key, value)
        count: int = len(steps)
        if count != 1:
            option = key + ("::" + value if value is not None else "")
            raise SapioException(f"{('No' if count == 0 else 'Multiple')} entries in this experiment match the "
                                 f"provided option: {option}")
        return steps[0]

    def get_steps_by_option(self, key: str, value: str | None = None) -> list[ElnEntryStep]:
        """
        Retrieve every step in this experiment that contains an entry option with the provided key and value.

        :param key: The key of the entry option to match on.
        :param value: The value of the entry option to match on. If not provided, then only matches on key.
        :return: The entries in this experiment that match the provided entry option key and value.
        """
        ret_list: list[ElnEntryStep] = []
        for step in self.get_all_steps():
            options: dict[str, str] = self.get_step_options(step)
            if key in options:
                if value is None or options[key] == value:
                    ret_list.append(step)
        return ret_list

    def get_step_records(self, step: Step) -> list[DataRecord]:
        """
        Query for the data records for the given step. The returned records are not cached by the ExperimentHandler.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step to get the data records for.
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :return: The data records for the given step.
        """
        return self.__to_eln_step(step).get_records()

    def get_step_models(self, step: Step, wrapper_type: type[WrappedType]) -> list[WrappedType]:
        """
        Query for the data records for the given step and wrap them as record models with the given type. The returned
        records are not cached by the ExperimentHandler.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step to get the data records for.
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :param wrapper_type: The record model wrapper to use.
        :return: The record models for the given step.
        """
        return self.__inst_man.add_existing_records_of_type(self.get_step_records(step), wrapper_type)

    def add_step_records(self, step: Step, records: Iterable[SapioRecord]) -> None:
        """
        Make a webservice call to add a list of records to a step. Only functions for global data type table entries.
        For adding to an ELN data type table entry, see add_eln_rows.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step to add the records to.
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :param records:
            A list of records to add to the given step.
            The records may be provided as either DataRecords, PyRecordModels, or WrappedRecordModels.
        """
        step = self.__to_eln_step(step)
        if not records:
            return
        dt: str = AliasUtil.to_singular_data_type_name(records)
        if ElnBaseDataType.is_base_data_type(dt):
            raise SapioException(f"{dt} is an ELN data type. This function call has no effect on ELN data types. "
                                 f"Use add_eln_rows or add_sample_details instead.")
        if dt != step.get_data_type_names()[0]:
            raise SapioException(f"Cannot add {dt} records to entry {step.get_name()} of type "
                                 f"{step.get_data_type_names()[0]}.")
        step.add_records(AliasUtil.to_data_records(records))

    def remove_step_records(self, step: Step, records: Iterable[SapioRecord]) -> None:
        """
        Make a webservice call to remove a list of records from a step. Only functions for global data type table
        entries. For removing from an ELN data type table entry, see remove_eln_rows.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :param records:
            A list of records to remove from the given step.
            The records may be provided as either DataRecords, PyRecordModels, or WrappedRecordModels.
        """
        step = self.__to_eln_step(step)
        if not records:
            return
        dt: str = AliasUtil.to_singular_data_type_name(records)
        if ElnBaseDataType.is_base_data_type(dt):
            raise SapioException(f"{dt} is an ELN data type. This function call has no effect on ELN data types. "
                                 f"Use remove_eln_rows or remove_sample_details instead.")
        if dt != step.get_data_type_names()[0]:
            raise SapioException(f"Cannot remove {dt} records from entry {step.get_name()} of type "
                                 f"{step.get_data_type_names()[0]}.")
        step.remove_records(AliasUtil.to_data_records(records))

    def set_step_records(self, step: Step, records: Iterable[SapioRecord]) -> None:
        """
        Sets the records in the given step to be equal to the input list of records. If a record is already on the step,
        it remains. If a record is missing from the step, it gets added. If a record is on the step but not in the
        provided record list, it gets removed. Makes one webservice call to get what is currently on the step and
        one additional webservice call for each of either adding or removing, if necessary.

        Functions for table, form, and attachment entries. For form and attachment entries, only a single record should
        be provided.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :param records:
            A list of records to set for the given step,
            The records may be provided as either DataRecords, PyRecordModels, or WrappedRecordModels.
        """
        step = self.__to_eln_step(step)
        if records:
            dt: str = AliasUtil.to_singular_data_type_name(records)
            if ElnBaseDataType.is_base_data_type(dt):
                raise SapioException(f"{dt} is an ELN data type. This function call has no effect on ELN data types. "
                                     f"Use add_eln_rows or add_sample_details instead.")
            if dt != step.get_data_type_names()[0]:
                raise SapioException(f"Cannot set {dt} records for entry {step.get_name()} of type "
                                     f"{step.get_data_type_names()[0]}.")
        step.set_records(AliasUtil.to_data_records(records))

    # FR-46496 - Provide alias of set_step_records for use with form entries.
    def set_form_record(self, step: Step, record: SapioRecord) -> None:
        """
        Sets the record for a form entry.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :param record:
            A record to set for the given step,
            The record may be provided as either a DataRecord, PyRecordModel, or WrappedRecordModel.
        """
        self.set_step_records(step, [record])

    # FR-46496 - Provide functions for adding and removing rows from an ELN data type entry.
    def add_eln_rows(self, step: Step, count: int, wrapper_type: type[WrappedType] | None = None) \
            -> list[PyRecordModel | WrappedType]:
        """
        Add rows to an ELNExperimentDetail or ELNSampleDetail table entry. The rows will not appear in the system
        until a record manager store and commit has occurred.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :param count: The number of new rows to add to the entry.
        :param wrapper_type: Optionally wrap the ELN data type in a record model wrapper. If not provided, returns
            an unwrapped PyRecordModel.
        :return: A list of the newly created rows.
        """
        step = self.__to_eln_step(step)
        if step.eln_entry.entry_type != ElnEntryType.Table:
            raise SapioException("The provided step is not a table entry.")
        dt: str = step.get_data_type_names()[0]
        if not ElnBaseDataType.is_eln_type(dt):
            raise SapioException("The provided step is not an ELN data type entry.")
        records: list[PyRecordModel] = self.__inst_man.add_new_records(dt, count)
        if wrapper_type:
            return self.__inst_man.wrap_list(records, wrapper_type)
        return records

    def add_eln_row(self, step: Step, wrapper_type: type[WrappedType] | None = None) -> PyRecordModel | WrappedType:
        """
        Add a row to an ELNExperimentDetail or ELNSampleDetail table entry. The row will not appear in the system
        until a record manager store and commit has occurred.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :param wrapper_type: Optionally wrap the ELN data type in a record model wrapper. If not provided, returns
            an unwrapped PyRecordModel.
        :return: The newly created row.
        """
        return self.add_eln_rows(step, 1, wrapper_type)[0]

    def remove_eln_rows(self, step: Step, records: list[SapioRecord]) -> None:
        """
        Remove rows from an ELNExperimentDetail or ELNSampleDetail table entry. ELN data type table entries display all
        records in the system that match the entry's data type. This means that removing rows from an ELN data type
        table entry is equivalent to deleting the records for the rows.

        The rows will not be deleted in the system until a record manager store and commit has occurred.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :param records:
            A list of records to remove from the given step.
            The records may be provided as either DataRecords, PyRecordModels, or WrappedRecordModels.
        """
        step = self.__to_eln_step(step)
        dt: str = step.get_data_type_names()[0]
        if not ElnBaseDataType.is_eln_type(dt):
            raise SapioException("The provided step is not an ELN data type entry.")
        if not records:
            return
        record_dt: str = AliasUtil.to_singular_data_type_name(records, False)
        if record_dt != dt:
            raise SapioException(f"Cannot remove {dt} records from entry {step.get_name()} of type "
                                 f"{step.get_data_type_names()[0]}.")
        # If any rows were provided as data records, turn them into record models before deleting them, as otherwise
        # this function would need to make a webservice call to do the deletion.
        data_records: list[DataRecord] = []
        for record in records:
            if isinstance(record, DataRecord):
                data_records.append(record)
            else:
                record.delete()
        if data_records:
            record_models: list[PyRecordModel] = self.__inst_man.add_existing_records(data_records)
            for record in record_models:
                record.delete()

    def remove_eln_row(self, step: Step, record: SapioRecord) -> None:
        """
        Remove a row from an ELNExperimentDetail or ELNSampleDetail table entry. ELN data type table entries display all
        records in the system that match the entry's data type. This means that removing rows from an ELN data type
        table entry is equivalent to deleting the records for the rows.

        The row will not be deleted in the system until a record manager store and commit has occurred.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :param record:
            The record to remove from the given step.
            The record may be provided as either a DataRecord, PyRecordModel, or WrappedRecordModel.
        """
        self.remove_eln_rows(step, [record])

    def add_sample_details(self, step: Step, samples: list[RecordModel], wrapper_type: type[WrappedType]) \
            -> list[PyRecordModel | WrappedType]:
        """
        Add sample details to a sample details entry while relating them to the input sample records.

        :param step:
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :param samples: The sample records to add the sample details to.
        :param wrapper_type: Optionally wrap the sample details in a record model wrapper. If not provided, returns
            an unwrapped PyRecordModel.
        :return: The newly created sample details. The indices of the samples in the input list match the index of the
            sample details in this list that they are related to.
        """
        step = self.__to_eln_step(step)
        if step.eln_entry.entry_type != ElnEntryType.Table:
            raise SapioException("The provided step is not a table entry.")
        dt: str = step.get_data_type_names()[0]
        if not ElnBaseDataType.is_eln_type(dt) or ElnBaseDataType.get_base_type(dt) != ElnBaseDataType.SAMPLE_DETAIL:
            raise SapioException("The provided step is not an ELNSampleDetail entry.")
        records: list[PyRecordModel] = []
        for sample in samples:
            if sample.data_type_name != "Sample":
                raise SapioException(f"Received a {sample.data_type_name} record when Sample records were expected.")
            detail: PyRecordModel = sample.add(Child.create_by_name(dt))
            detail.set_field_values({
                "SampleId": sample.get_field_value("SampleId"),
                "OtherSampleId": sample.get_field_value("OtherSampleId")
            })
            records.append(detail)
        if wrapper_type:
            return self.__inst_man.wrap_list(records, wrapper_type)
        return records

    def update_step(self, step: Step,
                    entry_name: str | None = None,
                    related_entry_set: Iterable[int] | None = None,
                    dependency_set: Iterable[int] | None = None,
                    entry_status: ExperimentEntryStatus | None = None,
                    order: int | None = None,
                    description: str | None = None,
                    requires_grabber_plugin: bool | None = None,
                    is_initialization_required: bool | None = None,
                    notebook_experiment_tab_id: int | None = None,
                    entry_height: int | None = None,
                    column_order: int | None = None,
                    column_span: int | None = None,
                    is_removable: bool | None = None,
                    is_renamable: bool | None = None,
                    source_entry_id: int | None = None,
                    clear_source_entry_id: bool | None = None,
                    is_hidden: bool | None = None,
                    is_static_View: bool | None = None,
                    is_shown_in_template: bool | None = None,
                    template_item_fulfilled_timestamp: int | None = None,
                    clear_template_item_fulfilled_timestamp: bool | None = None,
                    entry_options_map: dict[str, str] | None = None) -> None:
        """
        Make a webservice call to update an abstract step. If any parameter is not provided, then no change is made
        to it. All changes will be reflected by the ExperimentEntry of the Step that is being updated.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The entry step to update.
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :param entry_name: The new name of this entry.
        :param related_entry_set: The new set of entry IDs for the entries that are related (implicitly dependent) to
            this entry. Completely overwrites the existing related entries.
        :param dependency_set: The new set of entry IDs for the entries that are dependent (explicitly dependent) on
            this entry. Completely overwrites the existing dependent entries.
        :param entry_status: The new status of this entry.
        :param order: The row order of this entry in its tab.
        :param description: The new description of this entry.
        :param requires_grabber_plugin: Whether this entry's initialization is handled by a grabber plugin. If true,
            then is_initialization_required is forced to true by the server.
        :param is_initialization_required: Whether the user is required to manually initialize this entry.
        :param notebook_experiment_tab_id: The ID of the tab that this entry should appear on.
        :param entry_height: The height of this entry.
        :param column_order: The column order of this entry.
        :param column_span: How many columns this entry spans.
        :param is_removable: Whether this entry can be removed by the user.
        :param is_renamable: Whether this entry can be renamed by the user.
        :param source_entry_id: The ID of this entry from its template.
        :param clear_source_entry_id: True if the source entry ID should be cleared.
        :param is_hidden: Whether this entry is hidden from the user.
        :param is_static_View: Whether this entry is static. Static entries are uneditable and shared across all
            experiments of the same template.
        :param is_shown_in_template: Whether this entry is saved to and shown in the experiment's template.
        :param template_item_fulfilled_timestamp: A timestamp in milliseconds for when this entry was initialized.
        :param clear_template_item_fulfilled_timestamp: True if the template item fulfilled timestamp should be cleared,
            uninitializing the entry.
        :param entry_options_map:
            The new map of options for this entry. Completely overwrites the existing options map.
            Any changes to the entry options will update this ExperimentHandler's cache of entry options.
            If you wish to add options to the existing map of options that an entry has, use the
            add_step_options method.
        """
        step: ElnEntryStep = self.__to_eln_step(step)
        criteria = AbstractElnEntryUpdateCriteria(step.eln_entry.entry_type)

        # These two variables could be iterables that aren't lists. Convert them to plain
        # lists, since that's what the update criteria is expecting.
        if related_entry_set is not None:
            related_entry_set = list(related_entry_set)
        if dependency_set is not None:
            dependency_set = list(dependency_set)

        criteria.entry_name = entry_name
        criteria.related_entry_set = related_entry_set
        criteria.dependency_set = dependency_set
        criteria.entry_status = entry_status
        criteria.order = order
        criteria.description = description
        criteria.requires_grabber_plugin = requires_grabber_plugin
        criteria.is_initialization_required = is_initialization_required
        criteria.notebook_experiment_tab_id = notebook_experiment_tab_id
        criteria.entry_height = entry_height
        criteria.column_order = column_order
        criteria.column_span = column_span
        criteria.is_removable = is_removable
        criteria.is_renamable = is_renamable
        criteria.source_entry_id = source_entry_id
        criteria.clear_source_entry_id = clear_source_entry_id
        criteria.is_hidden = is_hidden
        criteria.is_static_View = is_static_View
        criteria.is_shown_in_template = is_shown_in_template
        criteria.template_item_fulfilled_timestamp = template_item_fulfilled_timestamp
        criteria.clear_template_item_fulfilled_timestamp = clear_template_item_fulfilled_timestamp
        criteria.entry_options_map = entry_options_map

        self.__eln_man.update_experiment_entry(self.__exp_id, step.get_id(), criteria)

        # Update the cached information for this entry in case it's needed by the caller after updating.
        entry: ExperimentEntry = step.eln_entry
        if entry_name is not None:
            # PR-46477 - Ensure that the previous name of the updated entry already existed in the cache.
            if entry.entry_name in self.__steps:
                self.__steps.pop(entry.entry_name)
            entry.entry_name = entry_name
            self.__steps.update({entry_name: step})
        if related_entry_set is not None:
            entry.related_entry_id_set = related_entry_set
        if dependency_set is not None:
            entry.dependency_set = dependency_set
        if entry_status is not None:
            entry.entry_status = entry_status
        if order is not None:
            entry.order = order
        if description is not None:
            entry.description = description
        if requires_grabber_plugin is not None:
            entry.requires_grabber_plugin = requires_grabber_plugin
        if is_initialization_required is not None:
            entry.is_initialization_required = is_initialization_required
        if notebook_experiment_tab_id is not None:
            entry.notebook_experiment_tab_id = notebook_experiment_tab_id
        if entry_height is not None:
            entry.entry_height = entry_height
        if column_order is not None:
            entry.column_order = column_order
        if column_span is not None:
            entry.column_span = column_span
        if is_removable is not None:
            entry.is_removable = is_removable
        if is_renamable is not None:
            entry.is_renamable = is_renamable
        if source_entry_id is not None:
            entry.source_entry_id = source_entry_id
        if clear_source_entry_id is True:
            entry.source_entry_id = None
        if is_hidden is not None:
            entry.is_hidden = is_hidden
        if is_static_View is not None:
            entry.is_static_View = is_static_View
        if is_shown_in_template is not None:
            entry.is_shown_in_template = is_shown_in_template
        if template_item_fulfilled_timestamp is not None:
            entry.template_item_fulfilled_timestamp = template_item_fulfilled_timestamp
        if clear_template_item_fulfilled_timestamp is True:
            entry.template_item_fulfilled_timestamp = None
        if entry_options_map is not None:
            self.__step_options.update({step.get_id(): entry_options_map})

    def get_step_option(self, step: Step, option: str) -> str:
        """
        Get the value of a specific entry option for the given step.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        Getting the step options requires a webservice query, which is made the first time any step option
        method is called for a specific step. The step options are cached so that subsequent calls of this
        method for that step don't make a webservice call.

        :param step:
            The step to check the options of.
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :param option: The entry option to search for.
        :return: The value of the input entry option for the input step.
        """
        return self.get_step_options(step).get(option)

    def get_step_options(self, step: Step) -> dict[str, str]:
        """
        Get the entire map of options for the input step.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        Getting the step options requires a webservice query, which is made the first time any step option
        method is called for any step in this experiment. The step options are cached so that subsequent calls of this
        method don't make a webservice call.

        :param step:
            The step to get the options of.
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :return: The map of options for the input step.
        """
        step = self.__to_eln_step(step)
        if step not in self.__step_options:
            self.__step_options.update(ExperimentReportUtil.get_experiment_entry_options(self.user, self.get_all_steps()))
        return self.__step_options[step.get_id()]

    def add_step_options(self, step: Step, mapping: Mapping[str, str]):
        """
        Add to the existing map of options for the input step. Makes a webservice call to update the step. Also
        updates the cache of the step's options.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        Getting the step options requires a webservice query, which is made the first time any step option
        method is called for a specific step. The step options are cached so that subsequent calls of this
        method for that step don't make a webservice call.

        :param step:
            The step to update the options of.
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :param mapping: The new options and values to add to the existing step options, provided as some Mapping
            (e.g. a Dict). If an option key already exists and is provided in the mapping, overwrites the existing value
            for that key.
        """
        options: dict[str, str] = self.get_step_options(step)
        options.update(mapping)
        self.update_step(step, entry_options_map=options)

    def initialize_step(self, step: Step) -> None:
        """
        Initialize the input step by setting its template item fulfilled timestamp to now. Makes a webservice call to
        update the step. Checks if the step already has a timestamp, and does nothing if so.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step to initialize.
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        """
        # Avoid unnecessary calls if the step is already initialized.
        step: ElnEntryStep = self.__to_eln_step(step)
        if step.eln_entry.template_item_fulfilled_timestamp is None:
            self.update_step(step, template_item_fulfilled_timestamp=round(time.time() * 1000))

    def uninitialize_step(self, step: Step) -> None:
        """
        Uninitialize the input step by clearing its template item fulfilled timestamp to now. Makes a webservice call to
        update the step. Checks if the step already doesn't have a timestamp, and does nothing if so.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step to uninitialize.
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        """
        # Avoid unnecessary calls if the step is already uninitialized.
        step: ElnEntryStep = self.__to_eln_step(step)
        if step.eln_entry.template_item_fulfilled_timestamp is not None:
            self.update_step(step, clear_template_item_fulfilled_timestamp=True)

    def complete_step(self, step: Step) -> None:
        """
        Submit the input step. Makes a webservice call to update the step. Checks if the step is already completed, and
        does nothing if so.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step to complete.
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        """
        step = self.__to_eln_step(step)
        if step.eln_entry.entry_status not in self.__ENTRY_COMPLETE_STATUSES:
            step.complete_step()
            step.eln_entry.entry_status = ExperimentEntryStatus.Completed

    def unlock_step(self, step: Step) -> None:
        """
        Set the status of the input step to UnlockedChangesRequired. Makes a webservice call to update the step. Checks
        if the step is already unlocked, and does nothing if so.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step to unlock.
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        """
        step = self.__to_eln_step(step)
        if step.eln_entry.entry_status in self.__ENTRY_LOCKED_STATUSES:
            step.unlock_step()
            step.eln_entry.entry_status = ExperimentEntryStatus.UnlockedChangesRequired

    def disable_step(self, step: Step) -> None:
        """
        Set the status of the input step to Disabled. This is the state that entries are in when they are waiting for
        entries that they are dependent upon to be submitted before they can be enabled. If you have unsubmitted an
        entry and want its dependent entries to be locked again, then you would use this to set their status to
        disabled.

        Makes a webservice call to update the step. Checks if the step is already unlocked, and does nothing if so.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step to disable.
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        """
        step = self.__to_eln_step(step)
        if step.eln_entry.entry_status in self.__ENTRY_LOCKED_STATUSES:
            self.update_step(step, entry_status=ExperimentEntryStatus.Disabled)

    def step_is_submitted(self, step: Step) -> bool:
        """
        Determine if the input step has already been submitted.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step to check.
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :return: True if the step's status is Completed or CompletedApproved. False otherwise.
        """
        return self.__to_eln_step(step).eln_entry.entry_status in self.__ENTRY_COMPLETE_STATUSES

    def step_is_locked(self, step: Step) -> bool:
        """
        Determine if the input step has been locked in any way.

        If no step functions have been called before and a step is being searched for by name, queries for the
        list of steps in the experiment and caches them.

        :param step:
            The step to check.
            The step may be provided as either a string for the name of the step or an ElnEntryStep.
            If given a name, throws an exception if no step of the given name exists in the experiment.
        :return: True if the step's status is Completed, CompletedApproved, Disabled, LockedAwaitingApproval,
            or LockedRejected. False otherwise.
        """
        return self.__to_eln_step(step).eln_entry.entry_status in self.__ENTRY_LOCKED_STATUSES

    def __to_eln_step(self, step: Step) -> ElnEntryStep:
        """
        Convert a variable that could be either a string or an ElnEntryStep to just an ElnEntryStep.
        This will query and cache the steps for the experiment if the input step is a name and the steps have not been
        cached before.

        :return: The input step as an ElnEntryStep.
        """
        return self.get_step(step) if isinstance(step, str) else step

    def __get_experiment_options(self) -> dict[str, str]:
        """
        Cache the experiment options if they haven't been cached yet.

        :return: The options for this experiment.
        """
        if hasattr(self, "_ExperimentHandler__exp_options"):
            return self.__exp_options
        self.__exp_options = self.__eln_man.get_notebook_experiment_options(self.__exp_id)
        return self.__exp_options
